import os

import requests_mock
import pytest

from audiomate import corpus
from audiomate.corpus import io
from audiomate.corpus.io import aed

from tests import resources
from . import reader_test as rt


@pytest.fixture
def downloader():
    return io.AEDDownloader()


@pytest.fixture()
def zip_data():
    with open(resources.get_resource_path(['sample_files', 'zip_sample_with_subfolder.zip']), 'rb') as f:
        return f.read()


class TestAEDDownloader:

    def test_download(self, zip_data, downloader, tmpdir):
        target_folder = tmpdir.strpath

        with requests_mock.Mocker() as mock:
            mock.get(aed.DATA_URL, content=zip_data)

            downloader.download(target_folder)

        assert os.path.isfile(os.path.join(target_folder, 'a.txt'))
        assert os.path.isfile(os.path.join(target_folder, 'subsub', 'b.txt'))
        assert os.path.isfile(os.path.join(target_folder, 'subsub', 'c.txt'))


class TestAEDReader(rt.CorpusReaderTest):

    SAMPLE_PATH = resources.sample_corpus_path('aed')
    FILE_TRACK_BASE_PATH = SAMPLE_PATH

    EXPECTED_NUMBER_OF_TRACKS = 10
    EXPECTED_TRACKS = [
        rt.ExpFileTrack('acoustic_guitar_16', os.path.join('test', 'acoustic_guitar_16.wav')),
        rt.ExpFileTrack('footstep_300', os.path.join('test', 'footstep_300.wav')),
        rt.ExpFileTrack('violin_36', os.path.join('test', 'violin_36.wav')),
        rt.ExpFileTrack('airplane_1', os.path.join('train', 'airplane', 'airplane_1.wav')),
        rt.ExpFileTrack('airplane_23', os.path.join('train', 'airplane', 'airplane_23.wav')),
        rt.ExpFileTrack('airplane_33', os.path.join('train', 'airplane', 'airplane_33.wav')),
        rt.ExpFileTrack('footstep_16', os.path.join('train', 'footstep', 'footstep_16.wav')),
        rt.ExpFileTrack('helicopter_9', os.path.join('train', 'helicopter', 'helicopter_9.wav')),
        rt.ExpFileTrack('tone_12', os.path.join('train', 'tone', 'tone_12.wav')),
        rt.ExpFileTrack('tone_35', os.path.join('train', 'tone', 'tone_35.wav')),
    ]

    EXPECTED_NUMBER_OF_ISSUERS = 0

    EXPECTED_NUMBER_OF_UTTERANCES = 10
    EXPECTED_UTTERANCES = [
        rt.ExpUtterance('acoustic_guitar_16', 'acoustic_guitar_16', None, 0, float('inf')),
        rt.ExpUtterance('footstep_300', 'footstep_300', None, 0, float('inf')),
        rt.ExpUtterance('violin_36', 'violin_36', None, 0, float('inf')),
        rt.ExpUtterance('airplane_1', 'airplane_1', None, 0, float('inf')),
        rt.ExpUtterance('airplane_23', 'airplane_23', None, 0, float('inf')),
        rt.ExpUtterance('airplane_33', 'airplane_33', None, 0, float('inf')),
        rt.ExpUtterance('footstep_16', 'footstep_16', None, 0, float('inf')),
        rt.ExpUtterance('helicopter_9', 'helicopter_9', None, 0, float('inf')),
        rt.ExpUtterance('tone_12', 'tone_12', None, 0, float('inf')),
        rt.ExpUtterance('tone_35', 'tone_35', None, 0, float('inf')),
    ]

    EXPECTED_LABEL_LISTS = {
        'airplane_23': [rt.ExpLabelList(corpus.LL_SOUND_CLASS, 1)],
        'tone_12': [rt.ExpLabelList(corpus.LL_SOUND_CLASS, 1)],
        'acoustic_guitar_16': [rt.ExpLabelList(corpus.LL_SOUND_CLASS, 1)],
    }

    EXPECTED_LABELS = {
        'airplane_23': [
            rt.ExpLabel(corpus.LL_SOUND_CLASS, 'airplane', 0, float('inf')),
        ],
        'tone_12': [
            rt.ExpLabel(corpus.LL_SOUND_CLASS, 'tone', 0, float('inf')),
        ],
        'acoustic_guitar_16': [
            rt.ExpLabel(corpus.LL_SOUND_CLASS, 'acoustic_guitar', 0, float('inf')),
        ],
    }

    EXPECTED_NUMBER_OF_SUBVIEWS = 2
    EXPECTED_SUBVIEWS = [
        rt.ExpSubview('train', [
            'airplane_1',
            'airplane_23',
            'airplane_33',
            'footstep_16',
            'helicopter_9',
            'tone_12',
            'tone_35',
        ]),
        rt.ExpSubview('test', [
            'acoustic_guitar_16',
            'footstep_300',
            'violin_36',
        ]),
    ]

    def load(self):
        reader = io.AEDReader()
        return reader.load(self.SAMPLE_PATH)



import os

from audiomate.corpus.io import downloader

import pytest
import requests_mock

from tests import resources


MOCK_URL = 'https://downloadings.dl/download'


@pytest.fixture()
def tar_data():
    path = resources.get_resource_path([
        'sample_files',
        'cv_corpus_v1.tar.gz'
    ])

    with open(path, 'rb') as f:
        return f.read()


@pytest.fixture()
def zip_data():
    path = resources.get_resource_path([
        'sample_files',
        'zip_sample_with_subfolder.zip'
    ])

    with open(path, 'rb') as f:
        return f.read()


class MockArchiveDownloader(downloader.ArchiveDownloader):

    def type(self):
        return 'mock'


class TestArchiveDownloader:

    def test_download_tar(self, tar_data, tmpdir):
        target_folder = tmpdir.strpath
        corpus_dl = MockArchiveDownloader(
            MOCK_URL,
            downloader.ArkType.TAR
        )

        with requests_mock.Mocker() as mock:
            mock.get(MOCK_URL, content=tar_data)
            corpus_dl.download(target_folder)

        base_folder = os.path.join(target_folder, 'common_voice')
        assert os.path.isdir(base_folder)

        assert os.path.isfile(os.path.join(base_folder, 'cv-valid-dev.csv'))
        assert os.path.isdir(os.path.join(base_folder, 'cv-valid-dev'))
        assert os.path.isfile(os.path.join(base_folder, 'cv-valid-train.csv'))
        assert os.path.isdir(os.path.join(base_folder, 'cv-valid-train'))

    def test_download_zip(self, zip_data, tmpdir):
        target_folder = tmpdir.strpath
        corpus_dl = MockArchiveDownloader(
            MOCK_URL,
            downloader.ArkType.ZIP
        )

        with requests_mock.Mocker() as mock:
            mock.get(MOCK_URL, content=zip_data)
            corpus_dl.download(target_folder)

        base_folder = os.path.join(target_folder, 'subfolder')
        assert os.path.isdir(base_folder)

        assert os.path.isfile(os.path.join(base_folder, 'a.txt'))
        assert os.path.isfile(os.path.join(base_folder, 'subsub', 'b.txt'))
        assert os.path.isfile(os.path.join(base_folder, 'subsub', 'c.txt'))

    def test_download_auto(self, zip_data, tmpdir):
        target_folder = tmpdir.strpath
        corpus_dl = MockArchiveDownloader(
            MOCK_URL,
            downloader.ArkType.AUTO
        )

        with requests_mock.Mocker() as mock:
            mock.get(MOCK_URL, content=zip_data)
            corpus_dl.download(target_folder)

        base_folder = os.path.join(target_folder, 'subfolder')
        assert os.path.isdir(base_folder)

        assert os.path.isfile(os.path.join(base_folder, 'a.txt'))
        assert os.path.isfile(os.path.join(base_folder, 'subsub', 'b.txt'))
        assert os.path.isfile(os.path.join(base_folder, 'subsub', 'c.txt'))

    def test_download_moves_files_up(self, zip_data, tmpdir):
        target_folder = tmpdir.strpath
        corpus_dl = MockArchiveDownloader(
            MOCK_URL,
            downloader.ArkType.AUTO,
            move_files_up=True
        )

        with requests_mock.Mocker() as mock:
            mock.get(MOCK_URL, content=zip_data)
            corpus_dl.download(target_folder)

        assert os.path.isfile(os.path.join(target_folder, 'a.txt'))
        assert os.path.isfile(os.path.join(target_folder, 'subsub', 'b.txt'))
        assert os.path.isfile(os.path.join(target_folder, 'subsub', 'c.txt'))



import abc
import copy


class Track(abc.ABC):
    """
    Track is the abstract base class for an audio track.

    Args:
        idx (str): A identifier to uniquely identify a track.
    """
    __slots__ = ['idx']

    def __init__(self, idx):
        self.idx = idx

    def __copy__(self):
        return Track(self.idx)

    def __deepcopy(self, memo):
        return copy.copy(self)

    @property
    @abc.abstractmethod
    def sampling_rate(self):
        """
        Return the sampling rate.
        """
        pass

    @property
    @abc.abstractmethod
    def num_channels(self):
        """
        Return the number of channels.
        """
        pass

    @property
    @abc.abstractmethod
    def num_samples(self):
        """
        Return the total number of samples.
        """
        pass

    @property
    @abc.abstractmethod
    def duration(self):
        """
        Return the duration in seconds.
        """
        pass

    @abc.abstractmethod
    def read_samples(self, sr=None, offset=0, duration=None):
        """
        Return the samples of the track.

        Args:
            sr (int): If ``None``, uses the native sampling-rate,
                      otherwise resamples to the given sampling rate.
            offset (float): The time in seconds, from where to start
                            reading the samples (rel. to the track start).
            duration (float): The length of the samples to read in seconds.

        Returns:
            np.ndarray: A numpy array containing the samples
            as a floating point (numpy.float32) time series.
        """
        pass

    @abc.abstractmethod
    def read_frames(self, frame_size, hop_size, offset=0,
                    duration=None, buffer_size=5760000):
        """
        Generator that reads and returns the samples of the track in frames.

        Args:
            frame_size (int): The number of samples per frame.
            hop_size (int): The number of samples between two frames.
            offset (float): The time in seconds, from where to start
                            reading the samples (rel. to the track start).
            duration (float): The length of the samples to read in seconds.

        Returns:
            Generator: A generator yielding a tuple for every frame.
            The first item is the frame,
            the second the sampling-rate and
            the third a boolean indicating if it is the last frame.
        """
        pass



import bisect
import math

import numpy as np

import audiomate
from audiomate import containers
from audiomate import processing
from . import iterator


class Dataset(object):
    """
    An abstract class representing a dataset. A dataset provides indexable access to data.
    An implementation of a concrete dataset should override the methods ``__len__`` and ``__getitem``.

    A sample returned from a dataset is a tuple containing the data for this sample from every container.
    The data from different containers is ordered in the way the containers were passed to the Dataset.

    Args:
        corpus_or_utt_ids (Corpus, list): Either a corpus or a list of utterances.
                                          This defines which utterances are considered for iterating.
        containers (list, Container): A single container or a list of containers.
    """

    def __init__(self, corpus_or_utt_ids, feature_containers):
        if isinstance(corpus_or_utt_ids, audiomate.corpus.CorpusView):
            self.utt_ids = sorted(list(corpus_or_utt_ids.utterances.keys()))
        else:
            self.utt_ids = sorted(corpus_or_utt_ids)

        if isinstance(feature_containers, containers.Container):
            self.containers = [feature_containers]
        else:
            self.containers = feature_containers

        if len(self.containers) == 0:
            raise ValueError('At least one container has to be provided!')

        for cnt in self.containers:
            if not Dataset.container_has_utterances(cnt, self.utt_ids):
                raise ValueError('Container {} does not contain all necessary utterances'.format(cnt.path))

    def __getitem__(self, item):
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError

    @staticmethod
    def container_has_utterances(container, keys):
        container_keys = set(container.keys())
        return container_keys.issuperset(keys)


class UtteranceDataset(Dataset):
    """
    A dataset wrapping data of complete utterances. Assuming two containers are given,
    a single sample contains all the data for an utterance from both containers.

    Args:
        corpus_or_utt_ids (Corpus, list): Either a corpus or a list of utterances.
                                          This defines which utterances are considered for iterating.
        source_containers (list, Container): A single container or a list of containers.
        return_length (bool): If True, the length of the data from every container is returned as well.
                              The length is appended after the data of every container
                              (e.g. [container1-data, container1-length, container2-data, container2-length])
        pad (bool): If True, samples that are shorter than the longest sample in the dataset are padded with 0.
                    The padding is done for each container individually.
                    (e.g. container1-data can be shorter than container2-data)
                    If padding is enabled, the lengths are always returned ``return_length = True``.
        transform (list, Processor): A single processor or a list of processors.
                                     Has to be the same number as the number of containers.
                                     If for some containers no transform should be applied,
                                     pass ``None``.

    Examples:

        >>> corpus = audiomate.Corpus.load('/path/to/corpus')
        >>> container_inputs = containers.FeatureContainer('/path/to/features.hdf5')
        >>> container_outputs = containers.Container('/path/to/targets.hdf5')
        >>>
        >>> ds = UtteranceDataset(corpus, [container_inputs, container_outputs])
        >>> len(ds) # Number of utterances/samples in the dataset
        114
        >>> ds[20] # Utterance/Sample (inputs, inputs-len, outputs, outputs-len) with index 20
        (
            array([[0.72991909, 0.20258683, 0.30574747, 0.53783217],
                   [0.38875413, 0.83611128, 0.49054591, 0.15710017],
                   [0.35153358, 0.40051009, 0.93647765, 0.29589257],
                   [0.97465772, 0.80160451, 0.81871436, 0.4892925 ],
                   [0.59310933, 0.8565602 , 0.95468696, 0.07933512]]),
            5,
            array([0, 9, 10, 4]),
            4
        )

    With ``pad=True`` the data is padded on the outermost dimension
    with zeros to match the length of the longest sequence/utterance.
    In the following it is assumed the longest sequence in the inputs is ``8`` and in the outputs ``6``.

        >>> corpus = audiomate.Corpus.load('/path/to/corpus')
        >>> container_inputs = containers.FeatureContainer('/path/to/features.hdf5')
        >>> container_outputs = containers.Container('/path/to/targets.hdf5')
        >>>
        >>> ds = UtteranceDataset(corpus, [container_inputs, container_outputs], pad=True)
        >>> len(ds) # Number of utterances/samples in the dataset
        114
        >>> ds[20] # Utterance/Sample (inputs, inputs-len, outputs, outputs-len) with index 20
        (
            array([[0.72991909, 0.20258683, 0.30574747, 0.53783217],
                   [0.38875413, 0.83611128, 0.49054591, 0.15710017],
                   [0.35153358, 0.40051009, 0.93647765, 0.29589257],
                   [0.97465772, 0.80160451, 0.81871436, 0.4892925 ],
                   [0.59310933, 0.8565602 , 0.95468696, 0.07933512],
                   [0.0, 0.0, 0.0, 0.0],
                   [0.0, 0.0, 0.0, 0.0],
                   [0.0, 0.0, 0.0, 0.0]
            ]),
            5,
            array([0, 9, 10, 4, 0, 0]),
            4
        )

    """

    def __init__(self, corpus_or_utt_ids, source_containers, return_length=True, pad=False, transform=None):
        super(UtteranceDataset, self).__init__(corpus_or_utt_ids, source_containers)

        self.pad = pad

        if transform is None:
            self.transform = []
        else:
            if isinstance(transform, processing.Processor):
                self.transform = [transform]
            else:
                self.transform = transform

            if len(self.transform) != len(self.containers):
                raise ValueError(
                    ('Number of transforms does not match number of containers!'
                     'Use None for containers without transform!')
                )

            for index, container_transform in enumerate(self.transform):
                if container_transform is not None and \
                        not isinstance(self.containers[index], containers.FeatureContainer):
                    raise ValueError(
                        ('Transforms can only be applied to'
                         'Feature-Containers (Transform at index {})').format(index))

        if self.pad:
            self.return_length = True
        else:
            self.return_length = return_length

        self.pad_lengths = self.longest_utterances_per_container()

    def __len__(self):
        return len(self.utt_ids)

    def __getitem__(self, item):
        utt_idx = self.utt_ids[item]

        sample = []

        for index, cnt in enumerate(self.containers):
            data = cnt.get(utt_idx, mem_map=False)

            if index < len(self.transform) and self.transform[index] is not None:
                data = self.transform[index].process_frames(
                    data,
                    cnt.sampling_rate,
                    offset=0,
                    last=True
                )

            size = data.shape[0]

            required_padded_length = self.pad_lengths[index]

            if self.pad and size < required_padded_length:
                pad_width = [(0, required_padded_length - size)]

                for i in range(1, len(data.shape)):
                    pad_width.append((0, 0))

                data = np.pad(data, pad_width, mode='constant', constant_values=0)

            sample.append(data)

            if self.return_length:
                sample.append(size)

        return sample

    def longest_utterances_per_container(self):
        """ Return a tuple/list containing the length of the longest utterance of ever container. """
        lengths = []

        for cnt in self.containers:
            longest_in_container = 0
            for utt_idx in self.utt_ids:
                utt_length = cnt._file[utt_idx].shape[0]
                longest_in_container = max(utt_length, longest_in_container)

            lengths.append(longest_in_container)

        return lengths


class MultiFrameDataset(Dataset):
    """
    A dataset wrapping chunks of frames of a corpus. A single sample represents a chunk of frames.

    A chunk doesn't overlap an utterances boundaries. So if the utterance length is not divisible by the chunk length,
    the last chunk of an utterance may be smaller than the chunk size.

    Args:
        corpus_or_utt_ids (Corpus, list): Either a corpus or a list of utterances.
                                          This defines which utterances are considered for iterating.
        containers (list, Container): A single container or a list of containers.
        frames_per_chunk (int): Number of subsequent frames in a single sample.
        return_length (bool): If True, the length of the chunk is returned as well. (default ``False``)
                              The length is appended to tuple as the last element.
                              (e.g. [container1-data, container2-data, length])
        pad (bool): If True, samples that are shorter are padded with zeros to match ``frames_per_chunk``.
                    If padding is enabled, the lengths are always returned ``return_length = True``.

    Note:
        For a multi-frame dataset it is expected that every container contains exactly one value/vector for every frame.
        So the first dimension of every array in every container have to match.

    Examples:

        >>> corpus = audiomate.Corpus.load('/path/to/corpus')
        >>> container_inputs = containers.FeatureContainer('/path/to/features.hdf5')
        >>> container_outputs = containers.Container('/path/to/targets.hdf5')
        >>>
        >>> ds = MultiFrameDataset(corpus, [container_inputs, container_outputs], 5)
        >>> len(ds) # Number of chunks in the dataset
        355
        >>> ds[20] # Chunk (inputs, outputs) with index 20
        (
            array([[0.72991909, 0.20258683, 0.30574747, 0.53783217],
                   [0.38875413, 0.83611128, 0.49054591, 0.15710017],
                   [0.35153358, 0.40051009, 0.93647765, 0.29589257],
                   [0.97465772, 0.80160451, 0.81871436, 0.4892925 ],
                   [0.59310933, 0.8565602 , 0.95468696, 0.07933512]]),
            array([[0.0, 1.0], [0.0, 1.0],[0.0, 1.0],[0.0, 1.0], [0.0, 1.0]])
        )

    If the length should be returned, pass ``True`` to ``return_length``
    (Except for chunks at the of utterances the length will be equal to ``frames_per_chunk``.)

        >>> corpus = audiomate.Corpus.load('/path/to/corpus')
        >>> container_inputs = containers.FeatureContainer('/path/to/features.hdf5')
        >>> container_outputs = containers.Container('/path/to/targets.hdf5')
        >>>
        >>> ds = MultiFrameDataset(corpus, [container_inputs, container_outputs], 5)
        >>> len(ds) # Number of chunks in the dataset
        355
        >>> ds[20] # Chunk (inputs, outputs) with index 20
        (
            array([[0.72991909, 0.20258683, 0.30574747, 0.53783217],
                   [0.38875413, 0.83611128, 0.49054591, 0.15710017],
                   [0.35153358, 0.40051009, 0.93647765, 0.29589257],
                   [0.97465772, 0.80160451, 0.81871436, 0.4892925 ],
                   [0.59310933, 0.8565602 , 0.95468696, 0.07933512]]),
            array([[0.0, 1.0], [0.0, 1.0],[0.0, 1.0],[0.0, 1.0], [0.0, 1.0]]),
            5
        )
    """

    def __init__(self, corpus_or_utt_ids, containers, frames_per_chunk, return_length=False, pad=False):
        super(MultiFrameDataset, self).__init__(corpus_or_utt_ids, containers)

        if frames_per_chunk < 1:
            raise ValueError('Chunk-size has to be at least 1!')

        self.frames_per_chunk = frames_per_chunk
        self.pad = pad

        if self.pad:
            self.return_length = True
        else:
            self.return_length = return_length

        self.regions = self.get_utt_regions()
        self.region_offsets = [x[0] for x in self.regions]

    def __len__(self):
        last_region = self.regions[-1]
        return last_region[0] + last_region[1]

    def __getitem__(self, item):
        # we search the region before the offset is higher than the item.
        region_index = bisect.bisect_right(self.region_offsets, item) - 1
        region = self.regions[region_index]

        frame_offset = (item - region[0]) * self.frames_per_chunk
        frame_end = frame_offset + self.frames_per_chunk

        data = [x[frame_offset:frame_end].astype(np.float32) for x in region[2]]
        size = data[0].shape[0]

        if self.pad and size < self.frames_per_chunk:

            padded_data = []
            for x in data:
                # Only pad the outermost (first) dimension
                pad_widths = [(0, 0)] * (len(x.shape) - 1)
                pad_widths.insert(0, (0, self.frames_per_chunk - size))
                padded_x = np.pad(x, pad_widths, mode='constant', constant_values=0)
                padded_data.append(padded_x)

            data = padded_data

        if self.return_length:
            data.append(size)

        return data

    def partitioned_iterator(self, partition_size, shuffle=True, seed=None):
        """
        Return a partitioning :class:`audiomate.feeding.MultiFrameIterator` for the dataset.

        Args:
            partition_size (str): Size of the partitions in bytes. The units ``k`` (kibibytes), ``m``
                                  (mebibytes) and ``g`` (gibibytes) are supported, i.e. a ``partition_size``
                                  of ``1g`` equates :math:`2^{30}` bytes.
            shuffle (bool): Indicates whether the data should be returned in
                            random order (``True``) or not (``False``).
            seed (int): Seed to be used for the random number generator.

        Returns:
            MultiFrameIterator: A partition iterator over the dataset.
        """
        return iterator.MultiFrameIterator(self.utt_ids, self.containers, partition_size, self.frames_per_chunk,
                                           return_length=self.return_length, pad=self.pad, shuffle=shuffle, seed=seed)

    def get_utt_regions(self):
        """
        Return the regions of all utterances, assuming all utterances are concatenated.
        It is assumed that the utterances are sorted in ascending order for concatenation.

        A region is defined by offset (in chunks), length (num-chunks) and
        a list of references to the utterance datasets in the containers.

        Returns:
            list: List of with a tuple for every utterances containing the region info.
        """

        regions = []
        current_offset = 0

        for utt_idx in sorted(self.utt_ids):
            offset = current_offset

            num_frames = []
            refs = []

            for cnt in self.containers:
                num_frames.append(cnt.get(utt_idx).shape[0])
                refs.append(cnt.get(utt_idx, mem_map=True))

            if len(set(num_frames)) != 1:
                raise ValueError('Utterance {} has not the same number of frames in all containers!'.format(utt_idx))

            num_chunks = math.ceil(num_frames[0] / float(self.frames_per_chunk))

            region = (offset, num_chunks, refs)
            regions.append(region)

            # Sets the offset for the next utterances
            current_offset += num_chunks

        return regions


class FrameDataset(MultiFrameDataset):
    """
    A dataset wrapping frames of a corpus. A single sample represents a single frame.

    Args:
        corpus_or_utt_ids (Corpus, list): Either a corpus or a list of utterances.
                                          This defines which utterances are considered for iterating.
        containers (list, Container): A single container or a list of containers.

    Note:
        For a frame dataset it is expected that every container contains exactly one value/vector for every frame.
        So the first dimension of every array in every container have to match.

    Example:
        >>> corpus = audiomate.Corpus.load('/path/to/corpus')
        >>> container_inputs = containers.FeatureContainer('/path/to/features.hdf5')
        >>> container_outputs = containers.Container('/path/to/targets.hdf5')
        >>>
        >>> ds = FrameDataset(corpus, [container_inputs, container_outputs])
        >>> len(ds) # Number of frames in the dataset
        2938
        >>> ds[293] # Frame (inputs, outputs) with index 293
        (
            array([0.58843831, 0.18128443, 0.19718328, 0.25284105]),
            array([0.0, 1.0])
        )
    """

    def __init__(self, corpus_or_utt_ids, containers):
        super(FrameDataset, self).__init__(corpus_or_utt_ids, containers, 1, return_length=False)

    def __getitem__(self, item):
        data = super(FrameDataset, self).__getitem__(item)

        # We have to remove the outermost dimension, which is 1 for chunk-size of 1 frame
        return [x[0] for x in data]

    def partitioned_iterator(self, partition_size, shuffle=True, seed=None):
        """
        Return a partitioning :class:`audiomate.feeding.FrameIterator` for the dataset.

        Args:
            partition_size (str): Size of the partitions in bytes. The units ``k`` (kibibytes), ``m``
                                  (mebibytes) and ``g`` (gibibytes) are supported, i.e. a ``partition_size``
                                  of ``1g`` equates :math:`2^{30}` bytes.
            shuffle (bool): Indicates whether the data should be returned in
                            random order (``True``) or not (``False``).
            seed (int): Seed to be used for the random number generator.

        Returns:
            FrameIterator: A partition iterator over the dataset.
        """
        return iterator.FrameIterator(self.utt_ids, self.containers, partition_size, shuffle=shuffle, seed=seed)



import os

import pytest
import requests_mock

from audiomate import corpus
from audiomate import issuers
from audiomate.corpus import io
from audiomate.corpus.io import tatoeba

from tests import resources
from . import reader_test as rt


@pytest.fixture()
def sample_audio_list_path():
    return resources.get_resource_path(['sample_corpora', 'tatoeba_download', 'sentences_with_audio.csv'])


@pytest.fixture()
def sample_sentence_list_path():
    return resources.get_resource_path(['sample_corpora', 'tatoeba_download', 'sentences.csv'])


@pytest.fixture()
def sample_audio_list_tar_bz():
    with open(resources.get_resource_path(['sample_corpora', 'tatoeba_download', 'sentences_with_audio.tar.bz2']),
              'rb') as f:
        return f.read()


@pytest.fixture()
def sample_sentence_list_tar_bz():
    with open(resources.get_resource_path(['sample_corpora', 'tatoeba_download', 'sentences.tar.bz2']), 'rb') as f:
        return f.read()


@pytest.fixture()
def sample_audio_content():
    with open(resources.get_resource_path(['wav_files', 'wav_2.wav']), 'rb') as f:
        return f.read()


class TestTatoebaDownloader:

    def test_load_audio_list(self, sample_audio_list_path):
        downloader = io.TatoebaDownloader()
        entries = downloader._load_audio_list(sample_audio_list_path)

        assert len(entries) == 5

        assert entries['247'] == ['gretelen', 'CC BY-NC 4.0', None]
        assert entries['1881'] == ['CK', 'CC BY-NC-ND 3.0', 'http://www.manythings.org/tatoeba']
        assert entries['6286'] == ['Phoenix', 'CC BY-NC 4.0', None]
        assert entries['2952354'] == ['pencil', 'CC BY-NC 4.0', None]
        assert entries['6921520'] == ['CK', 'CC BY-NC-ND 3.0', 'http://www.manythings.org/tatoeba']

    def test_load_audio_list_all(self, sample_audio_list_path):
        downloader = io.TatoebaDownloader(include_empty_licence=True)
        entries = downloader._load_audio_list(sample_audio_list_path)

        assert len(entries) == 7

        assert entries['141'] == ['BraveSentry', None, None]
        assert entries['247'] == ['gretelen', 'CC BY-NC 4.0', None]
        assert entries['1355'] == ['Nero', None, None]
        assert entries['1881'] == ['CK', 'CC BY-NC-ND 3.0', 'http://www.manythings.org/tatoeba']
        assert entries['6286'] == ['Phoenix', 'CC BY-NC 4.0', None]
        assert entries['2952354'] == ['pencil', 'CC BY-NC 4.0', None]
        assert entries['6921520'] == ['CK', 'CC BY-NC-ND 3.0', 'http://www.manythings.org/tatoeba']

    def test_load_audio_list_filter_license(self, sample_audio_list_path):
        downloader = io.TatoebaDownloader(include_licenses=['CC BY-NC 4.0'])
        entries = downloader._load_audio_list(sample_audio_list_path)

        assert len(entries) == 3

        assert entries['247'] == ['gretelen', 'CC BY-NC 4.0', None]
        assert entries['6286'] == ['Phoenix', 'CC BY-NC 4.0', None]
        assert entries['2952354'] == ['pencil', 'CC BY-NC 4.0', None]

    def test_load_sentence_list(self, sample_sentence_list_path):
        downloader = io.TatoebaDownloader()
        entries = downloader._load_sentence_list(sample_sentence_list_path)

        assert len(entries) == 8

        assert entries['141'] == ['eng', 'I want you to tell me why you did that.']
        assert entries['247'] == ['fra', 'Comment ça, je suis trop vieille pour ce poste ?']
        assert entries['511'] == ['deu', 'Wer will heiße Schokolade?']
        assert entries['524'] == ['deu', 'Das ist zu teuer!']
        assert entries['1355'] == ['epo', 'Mi panikis la homojn.']
        assert entries['6286'] == ['deu', 'Ich denke, ich habe genug gehört.']
        assert entries['299609'] == ['eng', 'He washes his car at least once a week.']
        assert entries['6921520'] == ['ita', 'Ho una zia che abita a Osaka.']

    def test_load_sentence_list_filter_languages(self, sample_sentence_list_path):
        downloader = io.TatoebaDownloader(include_languages=['deu', 'eng'])
        entries = downloader._load_sentence_list(sample_sentence_list_path)

        assert len(entries) == 5

        assert entries['141'] == ['eng', 'I want you to tell me why you did that.']
        assert entries['511'] == ['deu', 'Wer will heiße Schokolade?']
        assert entries['524'] == ['deu', 'Das ist zu teuer!']
        assert entries['6286'] == ['deu', 'Ich denke, ich habe genug gehört.']
        assert entries['299609'] == ['eng', 'He washes his car at least once a week.']

    def test_download(self, sample_audio_list_tar_bz, sample_sentence_list_tar_bz, sample_audio_content, tmpdir):
        downloader = io.TatoebaDownloader()

        with requests_mock.Mocker() as mock:
            mock.get(tatoeba.AUDIO_LIST_URL, content=sample_audio_list_tar_bz)
            mock.get(tatoeba.SENTENCE_LIST_URL, content=sample_sentence_list_tar_bz)

            mock.get('https://audio.tatoeba.org/sentences/eng/141.mp3', content=sample_audio_content)
            mock.get('https://audio.tatoeba.org/sentences/fra/247.mp3', content=sample_audio_content)
            mock.get('https://audio.tatoeba.org/sentences/epo/1355.mp3', content=sample_audio_content)
            mock.get('https://audio.tatoeba.org/sentences/deu/6286.mp3', content=sample_audio_content)
            mock.get('https://audio.tatoeba.org/sentences/ita/6921520.mp3', content=sample_audio_content)

            downloader.download(tmpdir.strpath)

            assert os.path.isfile(os.path.join(tmpdir.strpath, 'meta.txt'))

            assert not os.path.isfile(os.path.join(tmpdir.strpath, 'audio', 'eng', '141.mp3'))
            assert os.path.isfile(os.path.join(tmpdir.strpath, 'audio', 'fra', '247.mp3'))
            assert not os.path.isfile(os.path.join(tmpdir.strpath, 'audio', 'epo', '1355.mp3'))
            assert os.path.isfile(os.path.join(tmpdir.strpath, 'audio', 'deu', '6286.mp3'))
            assert os.path.isfile(os.path.join(tmpdir.strpath, 'audio', 'ita', '6921520.mp3'))

    def test_download_with_empty_licenses(self, sample_audio_list_tar_bz, sample_sentence_list_tar_bz,
                                          sample_audio_content, tmpdir):
        downloader = io.TatoebaDownloader(include_empty_licence=True)

        with requests_mock.Mocker() as mock:
            mock.get(tatoeba.AUDIO_LIST_URL, content=sample_audio_list_tar_bz)
            mock.get(tatoeba.SENTENCE_LIST_URL, content=sample_sentence_list_tar_bz)

            mock.get('https://audio.tatoeba.org/sentences/eng/141.mp3', content=sample_audio_content)
            mock.get('https://audio.tatoeba.org/sentences/fra/247.mp3', content=sample_audio_content)
            mock.get('https://audio.tatoeba.org/sentences/epo/1355.mp3', content=sample_audio_content)
            mock.get('https://audio.tatoeba.org/sentences/deu/6286.mp3', content=sample_audio_content)
            mock.get('https://audio.tatoeba.org/sentences/ita/6921520.mp3', content=sample_audio_content)

            downloader.download(tmpdir.strpath)

            assert os.path.isfile(os.path.join(tmpdir.strpath, 'meta.txt'))

            assert os.path.isfile(os.path.join(tmpdir.strpath, 'audio', 'eng', '141.mp3'))
            assert os.path.isfile(os.path.join(tmpdir.strpath, 'audio', 'fra', '247.mp3'))
            assert os.path.isfile(os.path.join(tmpdir.strpath, 'audio', 'epo', '1355.mp3'))
            assert os.path.isfile(os.path.join(tmpdir.strpath, 'audio', 'deu', '6286.mp3'))
            assert os.path.isfile(os.path.join(tmpdir.strpath, 'audio', 'ita', '6921520.mp3'))

    def test_download_with_filter_lang(self, sample_audio_list_tar_bz, sample_sentence_list_tar_bz,
                                       sample_audio_content, tmpdir):
        downloader = io.TatoebaDownloader(include_languages=['deu', 'eng'])

        with requests_mock.Mocker() as mock:
            mock.get(tatoeba.AUDIO_LIST_URL, content=sample_audio_list_tar_bz)
            mock.get(tatoeba.SENTENCE_LIST_URL, content=sample_sentence_list_tar_bz)

            mock.get('https://audio.tatoeba.org/sentences/eng/141.mp3', content=sample_audio_content)
            mock.get('https://audio.tatoeba.org/sentences/fra/247.mp3', content=sample_audio_content)
            mock.get('https://audio.tatoeba.org/sentences/epo/1355.mp3', content=sample_audio_content)
            mock.get('https://audio.tatoeba.org/sentences/deu/6286.mp3', content=sample_audio_content)
            mock.get('https://audio.tatoeba.org/sentences/ita/6921520.mp3', content=sample_audio_content)

            downloader.download(tmpdir.strpath)

            assert os.path.isfile(os.path.join(tmpdir.strpath, 'meta.txt'))

            assert not os.path.isfile(os.path.join(tmpdir.strpath, 'audio', 'eng', '141.mp3'))
            assert not os.path.isfile(os.path.join(tmpdir.strpath, 'audio', 'fra', '247.mp3'))
            assert not os.path.isfile(os.path.join(tmpdir.strpath, 'audio', 'epo', '1355.mp3'))
            assert os.path.isfile(os.path.join(tmpdir.strpath, 'audio', 'deu', '6286.mp3'))
            assert not os.path.isfile(os.path.join(tmpdir.strpath, 'audio', 'ita', '6921520.mp3'))

    def test_download_with_filter_license(self, sample_audio_list_tar_bz, sample_sentence_list_tar_bz,
                                          sample_audio_content, tmpdir):
        downloader = io.TatoebaDownloader(include_licenses=['CC BY-NC-ND 3.0'])

        with requests_mock.Mocker() as mock:
            mock.get(tatoeba.AUDIO_LIST_URL, content=sample_audio_list_tar_bz)
            mock.get(tatoeba.SENTENCE_LIST_URL, content=sample_sentence_list_tar_bz)

            mock.get('https://audio.tatoeba.org/sentences/eng/141.mp3', content=sample_audio_content)
            mock.get('https://audio.tatoeba.org/sentences/fra/247.mp3', content=sample_audio_content)
            mock.get('https://audio.tatoeba.org/sentences/epo/1355.mp3', content=sample_audio_content)
            mock.get('https://audio.tatoeba.org/sentences/deu/6286.mp3', content=sample_audio_content)
            mock.get('https://audio.tatoeba.org/sentences/ita/6921520.mp3', content=sample_audio_content)

            downloader.download(tmpdir.strpath)

            assert os.path.isfile(os.path.join(tmpdir.strpath, 'meta.txt'))

            assert not os.path.isfile(os.path.join(tmpdir.strpath, 'audio', 'eng', '141.mp3'))
            assert not os.path.isfile(os.path.join(tmpdir.strpath, 'audio', 'fra', '247.mp3'))
            assert not os.path.isfile(os.path.join(tmpdir.strpath, 'audio', 'epo', '1355.mp3'))
            assert not os.path.isfile(os.path.join(tmpdir.strpath, 'audio', 'deu', '6286.mp3'))
            assert os.path.isfile(os.path.join(tmpdir.strpath, 'audio', 'ita', '6921520.mp3'))


class TestTatoebaReader:

    SAMPLE_PATH = resources.sample_corpus_path('tatoeba')
    FILE_TRACK_BASE_PATH = os.path.join(SAMPLE_PATH, 'audio')

    EXPECTED_NUMBER_OF_TRACKS = 5
    EXPECTED_TRACKS = [
        rt.ExpFileTrack('141', os.path.join('eng', '141.mp3')),
        rt.ExpFileTrack('247', os.path.join('fra', '247.mp3')),
        rt.ExpFileTrack('1355', os.path.join('deu', '1355.mp3')),
        rt.ExpFileTrack('1881', os.path.join('deu', '1881.mp3')),
        rt.ExpFileTrack('6921520', os.path.join('eng', '6921520.mp3')),
    ]

    EXPECTED_NUMBER_OF_ISSUERS = 4
    EXPECTED_ISSUERS = [
        rt.ExpSpeaker('BraveSentry', 1, issuers.Gender.UNKNOWN, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('gretelen', 1, issuers.Gender.UNKNOWN, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('Nero', 1, issuers.Gender.UNKNOWN, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('CK', 2, issuers.Gender.UNKNOWN, issuers.AgeGroup.UNKNOWN, None),
    ]

    EXPECTED_NUMBER_OF_UTTERANCES = 5
    EXPECTED_UTTERANCES = [
        rt.ExpUtterance('141', '141', 'BraveSentry', 0, float('inf')),
        rt.ExpUtterance('247', '247', 'gretelen', 0, float('inf')),
        rt.ExpUtterance('1355', '1355', 'Nero', 0, float('inf')),
        rt.ExpUtterance('1881', '1881', 'CK', 0, float('inf')),
        rt.ExpUtterance('6921520', '6921520', 'CK', 0, float('inf')),
    ]

    EXPECTED_LABEL_LISTS = {
        '141': [rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT_RAW, 1)],
        '247': [rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT_RAW, 1)],
        '1355': [rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT_RAW, 1)],
        '1881': [rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT_RAW, 1)],
        '6921520': [rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT_RAW, 1)],
    }

    EXPECTED_LABELS = {
        '141': [
            rt.ExpLabel(
                corpus.LL_WORD_TRANSCRIPT, 'I want you to tell me why you did that.',
                0, float('inf')
            )
        ],
        '247': [
            rt.ExpLabel(
                corpus.LL_WORD_TRANSCRIPT, 'Comment ça, je suis trop vieille pour ce poste ?',
                0, float('inf')
            )
        ],
        '1355': [
            rt.ExpLabel(
                corpus.LL_WORD_TRANSCRIPT, 'Wer will heiße Schokolade?',
                0, float('inf')
            )
        ],
        '1881': [
            rt.ExpLabel(
                corpus.LL_WORD_TRANSCRIPT, 'Das ist zu teuer!',
                0, float('inf')
            )
        ],
        '6921520': [
            rt.ExpLabel(
                corpus.LL_WORD_TRANSCRIPT, 'He washes his car at least once a week.',
                0, float('inf')
            )
        ],

    }

    EXPECTED_NUMBER_OF_SUBVIEWS = 0

    def load(self):
        return io.TatoebaReader().load(self.SAMPLE_PATH)



import os

import numpy as np
import pytest

import audiomate
from audiomate import tracks
from audiomate import containers
from audiomate import annotations
from audiomate import issuers
from audiomate.corpus.subset import subview
from audiomate.corpus.io import MusanReader, KaldiWriter
from audiomate.corpus.io import UnknownWriterException, UnknownReaderException

from .. import resources


@pytest.fixture
def corpus():
    corpus = audiomate.Corpus()

    ex_file = tracks.FileTrack('existing_file', '../any/path.wav')
    ex_issuer = issuers.Issuer('existing_issuer')
    ex_utterance = tracks.Utterance('existing_utt', ex_file, issuer=ex_issuer)

    corpus.tracks['existing_file'] = ex_file
    corpus.issuers['existing_issuer'] = ex_issuer
    corpus.utterances['existing_utt'] = ex_utterance

    return corpus


class TestCorpus:

    #
    # TRACK ADD
    #

    def test_new_file(self, corpus):
        corpus.new_file('../some/path.wav', 'fid')

        assert corpus.num_tracks == 2
        assert corpus.tracks['fid'].idx == 'fid'
        assert corpus.tracks['fid'].path == os.path.abspath(
            os.path.join(os.getcwd(), '../some/path.wav')
        )

    def test_new_file_duplicate_idx(self, corpus):
        corpus.new_file('../some/other/path.wav', 'existing_file')

        assert corpus.num_tracks == 2
        assert corpus.tracks['existing_file_1'].idx == 'existing_file_1'
        assert corpus.tracks['existing_file_1'].path == os.path.abspath(
            os.path.join(os.getcwd(), '../some/other/path.wav')
        )

    def test_new_file_copy_file(self, corpus, tmpdir):
        file_path = resources.sample_wav_file('wav_1.wav')

        corpus.path = tmpdir.strpath
        corpus.new_file(file_path, 'fid', copy_file=True)

        assert corpus.num_tracks == 2
        assert corpus.tracks['fid'].path == os.path.join(tmpdir.strpath, 'files', 'fid.wav')

    def test_import_tracks(self, corpus):
        importing_tracks = [
            tracks.FileTrack('a', '/some/path.wav'),
            tracks.FileTrack('b', '/some/other/path.wav'),
            tracks.FileTrack('existing_file', '/some/otherer/path.wav'),
        ]

        idx_mapping = corpus.import_tracks(importing_tracks)

        assert corpus.num_tracks == 4

        assert 'a' in corpus.tracks.keys()
        assert corpus.tracks['a'].path == '/some/path.wav'

        assert 'b' in corpus.tracks.keys()
        assert corpus.tracks['b'].path == '/some/other/path.wav'

        assert 'existing_file_1' in corpus.tracks.keys()
        assert corpus.tracks['existing_file_1'].path == '/some/otherer/path.wav'

        assert len(idx_mapping) == 3
        assert 'a' in idx_mapping['a'].idx
        assert 'b' in idx_mapping['b'].idx
        assert idx_mapping['existing_file'].idx == 'existing_file_1'

    #
    #   UTT ADD
    #

    def test_new_utterance(self, corpus):
        corpus.new_utterance('some_utt', 'existing_file',
                             issuer_idx='existing_issuer', start=0, end=20)

        assert corpus.num_utterances == 2
        assert corpus.utterances['some_utt'].idx == 'some_utt'
        assert corpus.utterances['some_utt'].track.idx == 'existing_file'
        assert corpus.utterances['some_utt'].issuer.idx == 'existing_issuer'
        assert corpus.utterances['some_utt'].start == 0
        assert corpus.utterances['some_utt'].end == 20

    def test_new_utterance_duplicate_idx(self, corpus):
        corpus.new_utterance('existing_utt', 'existing_file',
                             issuer_idx='existing_issuer', start=0, end=20)

        assert corpus.num_utterances == 2
        assert corpus.utterances['existing_utt_1'].idx == 'existing_utt_1'
        assert corpus.utterances['existing_utt_1'].track.idx == 'existing_file'
        assert corpus.utterances['existing_utt_1'].issuer.idx == 'existing_issuer'
        assert corpus.utterances['existing_utt_1'].start == 0
        assert corpus.utterances['existing_utt_1'].end == 20

    def test_new_utterance_value_error_if_track_unknown(self, corpus):
        with pytest.raises(ValueError):
            corpus.new_utterance('some_utt', 'some_file', issuer_idx='iid', start=0, end=20)

    def test_import_utterances(self, corpus):
        importing_utterances = [
            tracks.Utterance('a', corpus.tracks['existing_file'],
                             corpus.issuers['existing_issuer'], 0, 10),
            tracks.Utterance('b', corpus.tracks['existing_file'],
                             corpus.issuers['existing_issuer'], 10, 20),
            tracks.Utterance('existing_utt', corpus.tracks['existing_file'],
                             corpus.issuers['existing_issuer'], 20, 30)
        ]

        mapping = corpus.import_utterances(importing_utterances)

        assert corpus.num_utterances == 4
        assert 'a' in corpus.utterances.keys()
        assert 'b' in corpus.utterances.keys()
        assert 'existing_utt_1' in corpus.utterances.keys()

        assert len(mapping) == 3
        assert mapping['a'].idx == 'a'
        assert mapping['b'].idx == 'b'
        assert mapping['existing_utt'].idx == 'existing_utt_1'

    def test_import_utterance_no_track(self, corpus):
        importing_utterances = [
            tracks.Utterance('a', tracks.FileTrack('notexist', 'notexist'),
                             corpus.issuers['existing_issuer'], 0, 10)
        ]

        with pytest.raises(ValueError):
            corpus.import_utterances(importing_utterances)

    def test_import_utterance_no_issuer(self, corpus):
        importing_utterances = [
            tracks.Utterance('a', corpus.tracks['existing_file'],
                             issuers.Issuer('notexist'), 0, 10)
        ]

        with pytest.raises(ValueError):
            corpus.import_utterances(importing_utterances)

    #
    #   ISSUER ADD
    #

    def test_new_issuer(self, corpus):
        corpus.new_issuer('some_iss', info={'hallo': 'velo'})

        assert corpus.num_issuers == 2
        assert corpus.issuers['some_iss'].idx == 'some_iss'
        assert corpus.issuers['some_iss'].info['hallo'] == 'velo'

    def test_new_issuer_duplicate_idx(self, corpus):
        corpus.new_issuer('existing_issuer', info={'hallo': 'velo'})

        assert corpus.num_issuers == 2
        assert corpus.issuers['existing_issuer_1'].idx == 'existing_issuer_1'
        assert corpus.issuers['existing_issuer_1'].info['hallo'] == 'velo'

    def test_import_issuers(self, corpus):
        importing_issuers = [
            issuers.Issuer('a'),
            issuers.Issuer('b'),
            issuers.Issuer('existing_issuer')
        ]

        mapping = corpus.import_issuers(importing_issuers)

        assert corpus.num_issuers == 4
        assert 'a' in corpus.issuers.keys()
        assert 'b' in corpus.issuers.keys()
        assert 'existing_issuer_1' in corpus.issuers.keys()

        assert len(mapping) == 3
        assert mapping['a'].idx == 'a'
        assert mapping['b'].idx == 'b'
        assert mapping['existing_issuer'].idx == 'existing_issuer_1'

    #
    #   FEAT CONT ADD
    #

    def test_new_feature_container(self, corpus, tmpdir):
        corpus.path = tmpdir.strpath
        corpus.new_feature_container('mfcc')

        assert corpus.num_feature_containers == 1
        assert corpus.feature_containers['mfcc'].path == os.path.join(
            tmpdir.strpath, 'features', 'mfcc'
        )

    #
    #   SUBVIEW ADD
    #
    def test_import_subview(self, corpus):
        train_set = subview.Subview(None, filter_criteria=[
            subview.MatchingUtteranceIdxFilter(utterance_idxs={'existing_utt'})
        ])

        corpus.import_subview('train', train_set)

        assert corpus.num_subviews == 1
        assert corpus.subviews['train'] == train_set
        assert corpus.subviews['train'].corpus == corpus

    #
    #   CREATION
    #

    def test_from_corpus(self):
        original = resources.create_dataset()
        copy = audiomate.Corpus.from_corpus(original)

        assert copy.num_tracks == 4
        assert copy.num_issuers == 3
        assert copy.num_utterances == 5
        assert copy.num_subviews == 2
        assert copy.num_feature_containers == 2

        original.tracks['wav-1'].path = '/changed/path.wav'
        assert original.tracks['wav-1'].path != copy.tracks['wav-1'].path

    def test_from_corpus_only_utterances_and_tracks(self):
        ds = audiomate.Corpus()
        ds.new_file('/random/path', 'file_1')
        ds.new_file('/random/path2', 'file_2')
        ds.new_utterance('utt_1', 'file_1')
        ds.new_utterance('utt_2', 'file_2')

        copy = audiomate.Corpus.from_corpus(ds)

        assert copy.num_tracks == 2
        assert copy.num_utterances == 2
        assert copy.num_issuers == 0

    #
    #    CORPUS READING
    #

    def test_load_throws_exception_when_reader_unknown(self):
        corpus = audiomate.Corpus()

        with pytest.raises(UnknownReaderException):
            corpus.load(resources.sample_corpus_path('default'), reader='does_not_exist')

    def test_load_with_default_reader_when_reader_unspecified(self):
        corpus = audiomate.Corpus()
        corpus = corpus.load(resources.sample_corpus_path('default'))

        assert corpus.name == 'default'
        assert corpus.path == resources.sample_corpus_path('default')
        assert corpus.num_tracks == 6
        assert 'file-1' in corpus.tracks
        assert 'file-2' in corpus.tracks
        assert 'file-3' in corpus.tracks
        assert 'file-4' in corpus.tracks

    def test_load_with_custom_reader_specified_by_name(self):
        corpus = audiomate.Corpus()
        corpus = corpus.load(resources.sample_corpus_path('musan'), reader='musan')

        assert corpus.name == 'musan'
        assert corpus.path == resources.sample_corpus_path('musan')
        assert corpus.num_tracks == 5
        assert 'music-fma-0000' in corpus.tracks
        assert 'noise-free-sound-0000' in corpus.tracks
        assert 'noise-free-sound-0001' in corpus.tracks
        assert 'speech-librivox-0000' in corpus.tracks
        assert 'speech-librivox-0001' in corpus.tracks

    def test_load_with_custom_reader_specified_by_instance(self):
        corpus = audiomate.Corpus()
        corpus = corpus.load(resources.sample_corpus_path('musan'), reader=MusanReader())

        assert corpus.name == 'musan'
        assert corpus.path == resources.sample_corpus_path('musan')
        assert corpus.num_tracks == 5
        assert 'music-fma-0000' in corpus.tracks
        assert 'noise-free-sound-0000' in corpus.tracks
        assert 'noise-free-sound-0001' in corpus.tracks
        assert 'speech-librivox-0000' in corpus.tracks
        assert 'speech-librivox-0001' in corpus.tracks

    #
    #    CORPUS SAVING
    #

    def test_save_at_corpus_path_throws_exception_when_writer_does_not_exist(self, tmpdir):
        corpus = audiomate.Corpus()
        corpus = corpus.load(resources.sample_corpus_path('default'))

        assert corpus.name == 'default'
        assert corpus.path == resources.sample_corpus_path('default')
        assert corpus.num_tracks == 6

        tempdir_contents = os.listdir(tmpdir.strpath)
        assert len(tempdir_contents) == 0

        corpus.path = tmpdir.strpath
        with pytest.raises(UnknownWriterException):
            corpus.save(writer='does_not_exist')

        assert len(os.listdir(tmpdir.strpath)) == 0

    def test_save_at_corpus_path_with_default_writer_when_writer_unspecified(self, tmpdir):
        corpus = audiomate.Corpus()
        corpus = corpus.load(resources.sample_corpus_path('default'))

        assert corpus.name == 'default'
        assert corpus.path == resources.sample_corpus_path('default')
        assert corpus.num_tracks == 6

        tempdir_contents = os.listdir(tmpdir.strpath)
        assert len(tempdir_contents) == 0

        corpus.path = tmpdir.strpath
        corpus.save()

        tempdir_contents = os.listdir(tmpdir.strpath)
        assert len(tempdir_contents) == 10

        assert 'files.txt' in tempdir_contents
        assert 'issuers.json' in tempdir_contents
        assert 'labels_raw_text.txt' in tempdir_contents
        assert 'labels_text.txt' in tempdir_contents
        assert 'utt_issuers.txt' in tempdir_contents
        assert 'utterances.txt' in tempdir_contents

    def test_save_at_corpus_path_with_writer_specified_by_name(self, tmpdir):
        corpus = audiomate.Corpus()
        corpus = corpus.load(resources.sample_corpus_path('kaldi'), reader='kaldi')

        assert corpus.name == 'kaldi'
        assert corpus.path == resources.sample_corpus_path('kaldi')
        assert corpus.path != tmpdir.strpath
        assert corpus.num_tracks == 4

        tempdir_contents = os.listdir(tmpdir.strpath)
        assert len(tempdir_contents) == 0

        corpus.path = tmpdir.strpath
        corpus.save(writer='kaldi')

        assert corpus.path == tmpdir.strpath

        tempdir_contents = os.listdir(tmpdir.strpath)
        assert len(tempdir_contents) == 4

        assert 'segments' in tempdir_contents
        assert 'text' in tempdir_contents
        assert 'utt2spk' in tempdir_contents
        assert 'wav.scp' in tempdir_contents

    def test_save_at_corpus_path_with_writer_specified_by_instance(self, tmpdir):
        corpus = audiomate.Corpus()
        corpus = corpus.load(resources.sample_corpus_path('kaldi'), reader='kaldi')

        assert corpus.name == 'kaldi'
        assert corpus.path == resources.sample_corpus_path('kaldi')
        assert corpus.path != tmpdir.strpath
        assert corpus.num_tracks == 4

        tempdir_contents = os.listdir(tmpdir.strpath)
        assert len(tempdir_contents) == 0

        corpus.path = tmpdir.strpath
        corpus.save(writer=KaldiWriter())

        assert corpus.path == tmpdir.strpath

        tempdir_contents = os.listdir(tmpdir.strpath)
        assert len(tempdir_contents) == 4

        assert 'segments' in tempdir_contents
        assert 'text' in tempdir_contents
        assert 'utt2spk' in tempdir_contents
        assert 'wav.scp' in tempdir_contents

    def test_save_at_path_throws_exception_when_writer_does_not_exist(self, tmpdir):
        corpus = audiomate.Corpus()
        corpus = corpus.load(resources.sample_corpus_path('default'))

        assert corpus.name == 'default'
        assert corpus.path == resources.sample_corpus_path('default')
        assert corpus.num_tracks == 6

        tempdir_contents = os.listdir(tmpdir.strpath)
        assert len(tempdir_contents) == 0

        with pytest.raises(UnknownWriterException):
            corpus.save_at(tmpdir.strpath, writer='does_not_exist')

        assert len(os.listdir(tmpdir.strpath)) == 0

    def test_save_at_path_with_default_writer_when_writer_unspecified(self, tmpdir):
        corpus = audiomate.Corpus()
        corpus = corpus.load(resources.sample_corpus_path('default'))

        assert corpus.name == 'default'
        assert corpus.path == resources.sample_corpus_path('default')
        assert corpus.num_tracks == 6

        tempdir_contents = os.listdir(tmpdir.strpath)
        assert len(tempdir_contents) == 0

        corpus.save_at(tmpdir.strpath)

        tempdir_contents = os.listdir(tmpdir.strpath)
        assert len(tempdir_contents) == 10

        assert 'files.txt' in tempdir_contents
        assert 'issuers.json' in tempdir_contents
        assert 'labels_raw_text.txt' in tempdir_contents
        assert 'labels_text.txt' in tempdir_contents
        assert 'utt_issuers.txt' in tempdir_contents
        assert 'utterances.txt' in tempdir_contents

    def test_save_at_path_with_writer_specified_by_name(self, tmpdir):
        corpus = audiomate.Corpus()
        corpus = corpus.load(resources.sample_corpus_path('kaldi'), reader='kaldi')

        assert corpus.name == 'kaldi'
        assert corpus.path == resources.sample_corpus_path('kaldi')
        assert corpus.path != tmpdir.strpath
        assert corpus.num_tracks == 4

        tempdir_contents = os.listdir(tmpdir.strpath)
        assert len(tempdir_contents) == 0

        corpus.save_at(tmpdir.strpath, writer='kaldi')

        assert corpus.path == tmpdir.strpath

        tempdir_contents = os.listdir(tmpdir.strpath)
        assert len(tempdir_contents) == 4

        assert 'segments' in tempdir_contents
        assert 'text' in tempdir_contents
        assert 'utt2spk' in tempdir_contents
        assert 'wav.scp' in tempdir_contents

    def test_save_at_path_with_writer_specified_by_instance(self, tmpdir):
        corpus = audiomate.Corpus()
        corpus = corpus.load(resources.sample_corpus_path('kaldi'), reader='kaldi')

        assert corpus.name == 'kaldi'
        assert corpus.path == resources.sample_corpus_path('kaldi')
        assert corpus.path != tmpdir.strpath
        assert corpus.num_tracks == 4

        tempdir_contents = os.listdir(tmpdir.strpath)
        assert len(tempdir_contents) == 0

        corpus.save_at(tmpdir.strpath, writer=KaldiWriter())

        assert corpus.path == tmpdir.strpath

        tempdir_contents = os.listdir(tmpdir.strpath)
        assert len(tempdir_contents) == 4

        assert 'segments' in tempdir_contents
        assert 'text' in tempdir_contents
        assert 'utt2spk' in tempdir_contents
        assert 'wav.scp' in tempdir_contents

    def test_merge_corpus_tracks(self):
        main_corpus = resources.create_dataset()
        merging_corpus = resources.create_multi_label_corpus()

        main_corpus.merge_corpus(merging_corpus)

        assert main_corpus.num_tracks == 8

        assert set(main_corpus.tracks.keys()) == {
            'wav-1', 'wav_2', 'wav_3', 'wav_4',
            'wav-1_1', 'wav_2_1', 'wav_3_1', 'wav_4_1'
        }

        assert main_corpus.tracks['wav-1_1'].idx == 'wav-1_1'
        assert main_corpus.tracks['wav-1_1'].path == merging_corpus.tracks['wav-1'].path

    def test_merge_corpus_issuers(self):
        main_corpus = resources.create_dataset()
        merging_corpus = resources.create_multi_label_corpus()

        main_corpus.merge_corpus(merging_corpus)

        assert main_corpus.num_issuers == 6

        assert set(main_corpus.issuers.keys()) == {
            'spk-1', 'spk-2', 'spk-3',
            'spk-1_1', 'spk-2_1', 'spk-3_1'
        }

        assert main_corpus.issuers['spk-1_1'].idx == 'spk-1_1'
        assert main_corpus.issuers['spk-1_1'].info == merging_corpus.issuers['spk-1'].info
        assert len(main_corpus.issuers['spk-1_1'].utterances) == 2

    def test_merge_corpus_utterances(self):
        main_corpus = resources.create_dataset()
        merging_corpus = resources.create_multi_label_corpus()

        main_corpus.merge_corpus(merging_corpus)

        assert main_corpus.num_utterances == 13

        assert set(main_corpus.utterances.keys()) == {
            'utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5',
            'utt-1_1', 'utt-2_1', 'utt-3_1', 'utt-4_1',
            'utt-5_1', 'utt-6', 'utt-7', 'utt-8'
        }

        assert main_corpus.utterances['utt-2_1'].track == main_corpus.tracks['wav_2_1']
        assert main_corpus.utterances['utt-2_1'].issuer == main_corpus.issuers['spk-1_1']
        assert main_corpus.utterances['utt-2_1'].start == merging_corpus.utterances['utt-2'].start
        assert main_corpus.utterances['utt-2_1'].end == merging_corpus.utterances['utt-2'].end
        assert main_corpus.utterances['utt-2_1'] in main_corpus.issuers['spk-1_1'].utterances

    def test_merge_corpus_label_lists(self):
        main_corpus = resources.create_dataset()
        merging_corpus = resources.create_multi_label_corpus()

        main_corpus.merge_corpus(merging_corpus)

        assert set(main_corpus.utterances['utt-2_1'].label_lists.keys()) == {'default'}

        ll = main_corpus.utterances['utt-2_1'].label_lists['default']

        assert ll == annotations.LabelList(labels=[
            annotations.Label('music', 0, 5),
            annotations.Label('speech', 5, 12),
            annotations.Label('music', 13, 15)
        ])

    def test_merge_corpus_subviews(self):
        main_corpus = resources.create_dataset()
        merging_corpus = resources.create_multi_label_corpus()

        main_corpus.merge_corpus(merging_corpus)

        assert main_corpus.num_subviews == 4

        assert main_corpus.subviews.keys() == {'train', 'dev', 'train_1', 'dev_1'}
        assert main_corpus.subviews['train_1'].corpus == main_corpus
        assert set(main_corpus.subviews['train_1'].filter_criteria[0].utterance_idxs) == {
            'utt-4_1', 'utt-5_1', 'utt-6'
        }

    def test_merge_corpus_feature_containers(self):
        main_corpus = resources.create_dataset()
        merging_corpus = resources.create_multi_label_corpus()

        main_corpus.merge_corpus(merging_corpus)

        assert main_corpus.num_feature_containers == 4

        main_feats = main_corpus.feature_containers
        merge_feats = merging_corpus.feature_containers

        assert set(main_feats.keys()) == {'mfcc', 'mel', 'mfcc_1', 'energy'}
        assert main_feats['mfcc_1'].path == merge_feats['mfcc'].path
        assert main_feats['energy'].path == merge_feats['energy'].path

    def test_merge_corpora(self):
        ds1 = resources.create_dataset()
        ds2 = resources.create_multi_label_corpus()
        ds3 = resources.create_single_label_corpus()

        ds = audiomate.Corpus.merge_corpora([ds1, ds2, ds3])

        assert ds.num_tracks == 12
        assert ds.num_utterances == 21
        assert ds.num_issuers == 9
        assert ds.num_subviews == 4
        assert ds.num_feature_containers == 4

    #
    # Varia
    #

    def test_relocate_audio_to_single_container(self, tmpdir):
        corpus = audiomate.Corpus.load(resources.sample_corpus_path('default'))

        target_container_path = os.path.join(tmpdir.strpath, 'audio')
        corpus.relocate_audio_to_single_container(target_container_path)

        assert os.path.isfile(target_container_path)

        cont = containers.AudioContainer(target_container_path)
        cont.open()

        assert cont.keys() == [
            'file-1',
            'file-2',
            'file-3',
            'file-4',
            'file-5',
            'file-6',
        ]

        assert corpus.tracks['file-1'].idx == 'file-1'
        assert corpus.tracks['file-1'].container.path == cont.path

        assert corpus.utterances['utt-1'].track.idx == 'file-1'
        assert corpus.utterances['utt-1'].track.container.path == cont.path

        cont.close()

    def test_relocate_audio_to_wav_files(self, tmpdir):
        old_corpus = audiomate.Corpus.load(resources.sample_corpus_path('default'))
        new_corpus = audiomate.Corpus.from_corpus(old_corpus)

        target_path = os.path.join(tmpdir.strpath, 'audio')
        new_corpus.relocate_audio_to_wav_files(target_path)

        assert os.path.isdir(target_path)

        old_track = old_corpus.tracks['file-1']
        new_track = new_corpus.tracks['file-1']
        assert new_track.path == os.path.join(target_path, 'file-1.wav')
        assert np.allclose(
            old_track.read_samples(),
            new_track.read_samples()
        )

        old_track = old_corpus.tracks['file-5']
        new_track = new_corpus.tracks['file-5']
        assert new_track.path == os.path.join(target_path, 'file-5.wav')
        assert np.allclose(
            old_track.read_samples(),
            new_track.read_samples()
        )



import numpy as np

from audiomate.processing import pipeline
from audiomate.processing.pipeline import base


class Multiply(pipeline.Computation):
    def __init__(self, factor, parent=None, name=None):
        super(Multiply, self).__init__(parent=parent, name=name)
        self.factor = factor

        self.frame_scale = 1.0
        self.hop_scale = 1.0

    def compute(self, chunk, sampling_rate, corpus=None, utterance=None):
        return chunk.data * self.factor

    def frame_transform_step(self, frame_size, hop_size):
        return frame_size * self.frame_scale, hop_size * self.hop_scale


class Add(pipeline.Computation):
    def __init__(self, value, parent=None, name=None):
        super(Add, self).__init__(parent=parent, name=name)
        self.value = value

        self.frame_scale = 1.0
        self.hop_scale = 1.0

    def compute(self, chunk, sampling_rate, corpus=None, utterance=None):
        return chunk.data + self.value

    def frame_transform_step(self, frame_size, hop_size):
        return frame_size * self.frame_scale, hop_size * self.hop_scale


class Concat(pipeline.Reduction):

    def __init__(self, parents, name=None, min_frames=1, left_context=0, right_context=0):
        super(Concat, self).__init__(parents, name=name, min_frames=min_frames,
                                     left_context=left_context, right_context=right_context)

        self.frame_scale = 1.0
        self.hop_scale = 1.0

    def compute(self, chunk, sampling_rate, corpus=None, utterance=None):
        return np.hstack(chunk.data)

    def frame_transform_step(self, frame_size, hop_size):
        return frame_size * self.frame_scale, hop_size * self.hop_scale


class StepDummy(pipeline.Computation):
    def __init__(self, parent=None, name=None, min_frames=0, left_context=0, right_context=0):
        super(StepDummy, self).__init__(parent=parent, name=name, min_frames=min_frames,
                                        left_context=left_context, right_context=right_context)

    def compute(self, chunk, sampling_rate, corpus=None, utterance=None):
        start = chunk.left_context
        end = chunk.data.shape[0] - chunk.right_context

        return chunk.data[start:end]


class TestStep:
    def test_process(self):
        add_a = Add(5)
        mul = Multiply(2, parent=add_a)
        add_b = Add(2)
        add_c = Add(3, parent=add_a)

        concat = Concat(parents=[mul, add_b, add_c])

        in_data = np.array([[0, 1, 2, 3]])
        out_data = concat.process_frames(in_data, 4)

        assert np.array_equal(out_data, np.array([[10, 12, 14, 16, 2, 3, 4, 5, 8, 9, 10, 11]]))

    def test_process_with_single_step_and_context(self):
        context = StepDummy(min_frames=2, left_context=2, right_context=1)

        out_data = context.process_frames(np.array([[0, 1, 2, 3]]), 4, offset=0, last=False)
        assert out_data is None

        out_data = context.process_frames(np.array([[4, 5, 6, 7]]), 4, offset=1, last=False)
        assert out_data is None

        out_data = context.process_frames(np.array([[8, 9, 10, 11]]), 4, offset=2, last=False)
        assert np.array_equal(out_data, np.array([[0, 1, 2, 3],
                                                  [4, 5, 6, 7]]))

        out_data = context.process_frames(np.array([[12, 13, 14, 15]]), 4, offset=3, last=True)
        assert np.array_equal(out_data, np.array([[8, 9, 10, 11],
                                                  [12, 13, 14, 15]]))

    def test_process_with_reduction_and_context(self):
        context = StepDummy(name='DummyContext', min_frames=2, left_context=2, right_context=1)
        add_a = Add(5, name='Add5')

        concat = Concat(parents=[context, add_a], name='Concat')

        out_data = concat.process_frames(np.array([[0, 1, 2, 3]]), 4, offset=0, last=False)
        assert out_data is None

        out_data = concat.process_frames(np.array([[4, 5, 6, 7]]), 4, offset=1, last=False)
        assert out_data is None

        out_data = concat.process_frames(np.array([[8, 9, 10, 11]]), 4, offset=2, last=False)
        assert np.array_equal(out_data, np.array([[0, 1, 2, 3, 5, 6, 7, 8],
                                                  [4, 5, 6, 7, 9, 10, 11, 12]]))

        out_data = concat.process_frames(np.array([[12, 13, 14, 15]]), 4, offset=3, last=True)
        assert np.array_equal(out_data, np.array([[8, 9, 10, 11, 13, 14, 15, 16],
                                                  [12, 13, 14, 15, 17, 18, 19, 20]]))

    def test_process_with_context(self):
        context = StepDummy(min_frames=2, left_context=2, right_context=1)
        add_a = Add(5)
        mul = Multiply(2, parent=add_a)
        add_b = Add(2, parent=context)
        add_c = Add(3, parent=add_a)

        concat = Concat(parents=[mul, add_b, add_c])

        out_data = concat.process_frames(np.array([[0, 1, 2, 3]]), 4, offset=0, last=False)
        assert out_data is None

        out_data = concat.process_frames(np.array([[4, 5, 6, 7]]), 4, offset=1, last=False)
        assert out_data is None

        out_data = concat.process_frames(np.array([[8, 9, 10, 11]]), 4, offset=2, last=False)
        assert np.array_equal(out_data, np.array([[10, 12, 14, 16, 2, 3, 4, 5, 8, 9, 10, 11],
                                                  [18, 20, 22, 24, 6, 7, 8, 9, 12, 13, 14, 15]]))

        out_data = concat.process_frames(np.array([[12, 13, 14, 15]]), 4, offset=3, last=True)
        assert np.array_equal(out_data, np.array([[26, 28, 30, 32, 10, 11, 12, 13, 16, 17, 18, 19],
                                                  [34, 36, 38, 40, 14, 15, 16, 17, 20, 21, 22, 23]]))

    def test_frame_transform(self):
        add_a = Add(5)
        mul = Multiply(2, parent=add_a)
        concat = Concat(parents=[add_a, mul])
        add_b = Add(3, parent=concat)

        add_a.frame_scale = 2.0
        add_a.hop_scale = 1.5

        add_b.frame_scale = 0.75
        add_b.hop_scale = 1.25

        tf_fs, tf_hs = add_b.frame_transform(200, 120)

        assert tf_fs == 300
        assert tf_hs == 225

    def test_frame_transform_not_mocked_step(self):
        add_a = Add(5)
        mul_a = StepDummy(parent=add_a)
        add_b = Add(3, parent=mul_a)
        concat = Concat(parents=[add_b, mul_a])

        mul_x = Multiply(3, parent=concat)
        mul_y = Multiply(4, parent=concat)

        mul_x.frame_scale = 2.0
        mul_x.hop_scale = 2.0

        mul_y.frame_scale = 2.0
        mul_y.hop_scale = 2.0

        concat_fin = Concat(parents=[mul_x, mul_y])

        tf_fs, tf_hs = concat_fin.frame_transform(200, 120)

        assert tf_fs == 400
        assert tf_hs == 240


class TestBuffer:

    def test_not_enough_frames_and_not_last_returns_none(self):
        buffer = base.Buffer(3, 0, 0)
        chunk = np.arange(4).reshape(2, 2)

        buffer.update(chunk, 0, False)
        res = buffer.get()

        assert res is None

    def test_not_enough_frames_and_is_last_returns_rest(self):
        buffer = base.Buffer(3, 0, 0)
        chunk = np.arange(4).reshape(2, 2)
        buffer.update(chunk, 0, True)
        res = buffer.get()

        assert np.allclose(chunk, res.data)
        assert res.offset == 0
        assert res.is_last
        assert res.left_context == 0
        assert res.right_context == 0

    def test_has_enough_returns_all(self):
        buffer = base.Buffer(3, 0, 0)
        chunk1 = np.arange(4).reshape(2, 2)
        chunk2 = np.arange(6).reshape(3, 2) + 4

        buffer.update(chunk1, 0, False)
        res = buffer.get()
        assert res is None

        buffer.update(chunk2, 2, False)
        res = buffer.get()
        assert np.allclose(np.vstack([chunk1, chunk2]), res.data)
        assert res.offset == 0
        assert not res.is_last
        assert res.left_context == 0
        assert res.right_context == 0

    def test_returns_correct_offset(self):
        buffer = base.Buffer(2, 0, 0)
        chunk1 = np.arange(4).reshape(2, 2)
        chunk2 = np.arange(6).reshape(3, 2) + 4

        buffer.update(chunk1, 0, False)
        res = buffer.get()
        assert res.offset == 0

        buffer.update(chunk2, 2, False)
        res = buffer.get()
        assert res.offset == 2

    def test_with_left_context(self):
        buffer = base.Buffer(2, 3, 0)
        chunk = np.arange(4).reshape(2, 2)

        buffer.update(chunk, 0, False)
        res = buffer.get()
        assert np.allclose(chunk, res.data)
        assert res.offset == 0
        assert not res.is_last
        assert res.left_context == 0
        assert res.right_context == 0

    def test_returns_correct_number_of_frames_with_left_context(self):
        buffer = base.Buffer(2, 3, 0)
        chunk1 = np.arange(8).reshape(4, 2)
        chunk2 = np.arange(6).reshape(3, 2) + 8

        buffer.update(chunk1, 0, False)
        res = buffer.get()
        assert np.allclose(chunk1, res.data)
        assert res.offset == 0
        assert not res.is_last
        assert res.left_context == 0
        assert res.right_context == 0

        buffer.update(chunk2, 4, False)
        res = buffer.get()
        assert np.allclose(np.vstack([chunk1[-3:], chunk2]), res.data)
        assert res.offset == 1
        assert not res.is_last
        assert res.left_context == 3
        assert res.right_context == 0

    def test_returns_correct_number_of_left_context(self):
        buffer = base.Buffer(2, 3, 0)
        chunk1 = np.arange(4).reshape(2, 2)
        chunk2 = np.arange(6).reshape(3, 2) + 4

        buffer.update(chunk1, 0, False)
        res = buffer.get()
        assert res.offset == 0
        assert not res.is_last
        assert res.left_context == 0
        assert res.right_context == 0

        buffer.update(chunk2, 2, False)
        res = buffer.get()
        assert res.offset == 0
        assert not res.is_last
        assert res.left_context == 2
        assert res.right_context == 0

    def test_dont_include_context_into_min_frames(self):
        buffer = base.Buffer(3, 1, 0)
        chunk1 = np.arange(8).reshape(4, 2)
        chunk2 = np.arange(4).reshape(2, 2) + 8
        chunk3 = np.arange(2).reshape(1, 2) + 11

        buffer.update(chunk1, 0, False)
        res = buffer.get()
        assert res.data.shape[0] == 4

        buffer.update(chunk2, 4, False)
        res = buffer.get()
        assert res is None

        buffer.update(chunk3, 6, False)
        res = buffer.get()
        assert res.data.shape[0] == 4
        assert res.offset == 3
        assert not res.is_last
        assert res.left_context == 1
        assert res.right_context == 0

    def test_not_enough_frames_and_is_last_returns_rest_with_left_context(self):
        buffer = base.Buffer(3, 2, 0)

        chunk1 = np.arange(8).reshape(4, 2)
        chunk2 = np.arange(4).reshape(2, 2) + 8

        buffer.update(chunk1, 0, False)
        res = buffer.get()
        assert np.allclose(chunk1, res.data)
        assert res.offset == 0
        assert not res.is_last
        assert res.left_context == 0
        assert res.right_context == 0

        buffer.update(chunk2, 4, True)
        res = buffer.get()
        assert np.allclose(np.vstack([chunk1[2:], chunk2]), res.data)
        assert res.offset == 2
        assert res.is_last
        assert res.left_context == 2
        assert res.right_context == 0

    def test_right_context(self):
        buffer = base.Buffer(2, 1, 2)

        chunk1 = np.arange(4).reshape(2, 2)
        chunk2 = np.arange(6).reshape(3, 2) + 4

        buffer.update(chunk1, 0, False)
        res = buffer.get()
        assert res is None

        buffer.update(chunk2, 2, False)
        res = buffer.get()
        assert np.allclose(np.vstack([chunk1, chunk2]), res.data)
        assert res.offset == 0
        assert not res.is_last
        assert res.left_context == 0
        assert res.right_context == 2

    def test_not_enough_frames_and_is_last_returns_rest_with_right_context(self):
        buffer = base.Buffer(10, 1, 2)

        chunk1 = np.arange(24).reshape(12, 2)
        chunk2 = np.arange(2).reshape(1, 2) + 24

        buffer.update(chunk1, 0, False)
        res = buffer.get()
        assert np.allclose(chunk1, res.data)
        assert res.offset == 0
        assert not res.is_last
        assert res.left_context == 0
        assert res.right_context == 2

        buffer.update(chunk2, 12, True)
        res = buffer.get()
        assert np.allclose(np.vstack([chunk1[9:], chunk2]), res.data)
        assert res.offset == 9
        assert res.is_last
        assert res.left_context == 1
        assert res.right_context == 0

    def test_multiple_buffers(self):
        buffer = base.Buffer(3, 2, 0, num_buffers=2)

        chunk1 = np.arange(8).reshape(4, 2)
        buffer.update(chunk1, 0, False, buffer_index=0)
        buffer.update(chunk1, 0, False, buffer_index=1)
        res = buffer.get()

        assert len(res.data) == 2
        assert np.allclose(chunk1, res.data[0])
        assert np.allclose(chunk1, res.data[1])
        assert res.offset == 0
        assert not res.is_last
        assert res.left_context == 0
        assert res.right_context == 0

        chunk2 = np.arange(4).reshape(2, 2) + 8
        buffer.update(chunk2, 4, True, buffer_index=0)
        res = buffer.get()

        assert res is None

        buffer.update(chunk2, 4, True, buffer_index=1)
        res = buffer.get()

        assert len(res.data) == 2
        assert np.allclose(np.vstack([chunk1[2:], chunk2]), res.data[0])
        assert np.allclose(np.vstack([chunk1[2:], chunk2]), res.data[1])
        assert res.offset == 2
        assert res.is_last
        assert res.left_context == 2
        assert res.right_context == 0

    def test_multiple_buffers_with_different_income_timings(self):
        buffer = base.Buffer(min_frames=1, left_context=0, right_context=0, num_buffers=2)

        buffer.update(np.array([[1, 2, 3, 4]]), 0, False, buffer_index=1)
        res = buffer.get()
        assert res is None

        buffer.update(np.array([[5, 6, 7, 8]]), 1, False, buffer_index=1)
        res = buffer.get()
        assert res is None

        buffer.update(np.array([[0, 1, 2, 3], [4, 5, 6, 7]]), 0, False, buffer_index=0)
        buffer.update(np.array([[9, 10, 11, 12]]), 2, False, buffer_index=1)
        res = buffer.get()

        assert buffer.buffers[0].shape[0] == 0
        assert np.array_equal(buffer.buffers[1], np.array([[9, 10, 11, 12]]))

        assert len(res.data) == 2
        assert np.array_equal(res.data[0], np.array([[0, 1, 2, 3], [4, 5, 6, 7]]))
        assert np.array_equal(res.data[1], np.array([[1, 2, 3, 4], [5, 6, 7, 8]]))
        assert not res.is_last
        assert res.offset == 0
        assert res.left_context == 0
        assert res.right_context == 0

        buffer.update(np.array([[8, 9, 10, 11], [12, 13, 14, 15]]), 2, True, buffer_index=0)
        buffer.update(np.array([[13, 14, 15, 16]]), 3, True, buffer_index=1)
        res = buffer.get()

        assert len(res.data) == 2
        assert np.array_equal(res.data[0], np.array([[8, 9, 10, 11], [12, 13, 14, 15]]))
        assert np.array_equal(res.data[1], np.array([[9, 10, 11, 12], [13, 14, 15, 16]]))
        assert res.is_last
        assert res.offset == 2
        assert res.left_context == 0
        assert res.right_context == 0



import os

from audiomate.formats import trn


class TestTrnFormat:

    def test_write(self, tmpdir):
        path = os.path.join(tmpdir.strpath, 'transcriptions.txt')
        entries = {
            'utt-2': 'other text',
            'utt-1': 'some text',
        }

        trn.write(path, entries)

        with open(path, 'r') as f:
            lines = f.readlines()

        assert lines[0] == 'some text (utt-1)\n'
        assert lines[1] == 'other text (utt-2)'

    def test_read(self):
        path = os.path.join(os.path.dirname(__file__), 'trn_transcriptions.txt')
        entries = trn.read(path)

        assert entries['utt-2'] == 'hallo du'
        assert entries['utt-3'] == 'wer ist da'
        assert entries['utt-5'] == 'wie wer (oder wo) ist das'



import pytest

from audiomate import containers
from tests import resources


@pytest.fixture()
def sample_feature_container():
    container_path = resources.get_resource_path(
        ['sample_files', 'feat_container']
    )
    sample_container = containers.FeatureContainer(container_path)
    sample_container.open()
    yield sample_container
    sample_container.close()


class TestFeatureContainer:

    def test_frame_size(self, sample_feature_container):
        assert sample_feature_container.frame_size == 400

    def test_hop_size(self, sample_feature_container):
        assert sample_feature_container.hop_size == 160

    def test_sampling_rate(self, sample_feature_container):
        assert sample_feature_container.sampling_rate == 16000

    def test_stats_per_key(self, sample_feature_container):
        utt_stats = sample_feature_container.stats_per_key()

        assert utt_stats['utt-1'].min == pytest.approx(0.0071605651933048797)
        assert utt_stats['utt-1'].max == pytest.approx(0.9967182746569494)
        assert utt_stats['utt-1'].mean == pytest.approx(0.51029100520776705)
        assert utt_stats['utt-1'].var == pytest.approx(0.079222738766221268)
        assert utt_stats['utt-1'].num == 100

        assert utt_stats['utt-2'].min == pytest.approx(0.01672865642756316)
        assert utt_stats['utt-2'].max == pytest.approx(0.99394433783429104)
        assert utt_stats['utt-2'].mean == pytest.approx(0.46471979908661543)
        assert utt_stats['utt-2'].var == pytest.approx(0.066697466410977804)
        assert utt_stats['utt-2'].num == 65

        assert utt_stats['utt-3'].min == pytest.approx(0.014999482706963607)
        assert utt_stats['utt-3'].max == pytest.approx(0.99834417857609881)
        assert utt_stats['utt-3'].mean == pytest.approx(0.51042690965262705)
        assert utt_stats['utt-3'].var == pytest.approx(0.071833200069641057)
        assert utt_stats['utt-3'].num == 220

    def test_stats_per_key_not_open(self, sample_feature_container):
        sample_feature_container.close()

        with pytest.raises(ValueError):
            sample_feature_container.stats_per_key()

    def test_stats(self, sample_feature_container):
        stats = sample_feature_container.stats()

        assert stats.min == pytest.approx(0.0071605651933048797)
        assert stats.max == pytest.approx(0.99834417857609881)
        assert stats.mean == pytest.approx(0.50267482489606408)
        assert stats.var == pytest.approx(0.07317811077366114)

    def test_stats_not_open(self, sample_feature_container):
        sample_feature_container.close()

        with pytest.raises(ValueError):
            sample_feature_container.stats()



import copy

import librosa
import numpy as np

from . import track


class ContainerTrack(track.Track):
    """
    A track that is stored in a :py:class:`audiomate.containers.AudioContainer`.

    Args:
        idx (str): A identifier to uniquely identify a track.
        container (AudioContainer): The audio container with the samples.
        key (str): The key of the samples in the container.
                   If ``None``, it is assumed it's the same
                   as ``idx``.
    """
    __slots__ = ['container', 'key']

    def __init__(self, idx, container, key=None):
        super(ContainerTrack, self).__init__(idx)

        self.container = container

        if key is None:
            self.key = idx
        else:
            self.key = key

    def __copy__(self):
        return ContainerTrack(
            self.idx,
            self.container,
            key=self.key
        )

    def __deepcopy(self, memo):
        return ContainerTrack(
            self.idx,
            copy.deepcopy(self.container, memo),
            key=self.key
        )

    @property
    def sampling_rate(self):
        """
        Return the sampling rate.
        """
        with self.container.open_if_needed(mode='r') as cnt:
            return cnt.get(self.key)[1]

    @property
    def num_channels(self):
        """
        Return the number of channels.
        """
        return 1

    @property
    def num_samples(self):
        """
        Return the total number of samples.
        """
        with self.container.open_if_needed(mode='r') as cnt:
            return cnt.get(self.key)[0].shape[0]

    @property
    def duration(self):
        """
        Return the duration in seconds.
        """
        with self.container.open_if_needed(mode='r') as cnt:
            samples, sr = cnt.get(self.key)

            return samples.shape[0] / sr

    def read_samples(self, sr=None, offset=0, duration=None):
        """
        Return the samples from the track in the container.
        Uses librosa for resampling, if needed.

        Args:
            sr (int): If ``None``, uses the sampling rate given by the file,
                      otherwise resamples to the given sampling rate.
            offset (float): The time in seconds, from where to start reading
                            the samples (rel. to the file start).
            duration (float): The length of the samples to read in seconds.

        Returns:
            np.ndarray: A numpy array containing the samples as a
            floating point (numpy.float32) time series.
        """
        with self.container.open_if_needed(mode='r') as cnt:
            samples, native_sr = cnt.get(self.key)

            start_sample_index = int(offset * native_sr)

            if duration is None:
                end_sample_index = samples.shape[0]
            else:
                end_sample_index = int((offset + duration) * native_sr)

            samples = samples[start_sample_index:end_sample_index]

            if sr is not None and sr != native_sr:
                samples = librosa.core.resample(
                    samples,
                    native_sr,
                    sr,
                    res_type='kaiser_best'
                )

            return samples

    def read_frames(self, frame_size, hop_size, offset=0,
                    duration=None, block_size=None):
        """
        Generator that reads and returns the samples of the track in frames.

        Args:
            frame_size (int): The number of samples per frame.
            hop_size (int): The number of samples between two frames.
            offset (float): The time in seconds, from where to start
                            reading the samples (rel. to the track start).
            duration (float): The length of the samples to read in seconds.

        Returns:
            Generator: A generator yielding a tuple for every frame.
            The first item is the frame,
            the second the sampling-rate and
            the third a boolean indicating if it is the last frame.
        """
        with self.container.open_if_needed(mode='r') as cnt:
            samples, sr = cnt.get(self.key)

            current_index = 0

            while current_index + frame_size < samples.shape[0]:
                next_frame = samples[current_index:current_index+frame_size]
                yield next_frame, False
                current_index += hop_size

            next_frame = samples[current_index:]

            if next_frame.shape[0] < frame_size:
                next_frame = np.pad(
                    next_frame,
                    (0, frame_size - next_frame.shape[0]),
                    mode='constant',
                    constant_values=0
                )

            yield next_frame, True



import os

from audiomate import corpus
from audiomate.corpus import io
from audiomate import issuers

from tests import resources
from . import reader_test as rt


class TestTimitReader(rt.CorpusReaderTest):

    SAMPLE_PATH = resources.sample_corpus_path('timit')
    FILE_TRACK_BASE_PATH = SAMPLE_PATH

    EXPECTED_NUMBER_OF_TRACKS = 9
    EXPECTED_TRACKS = [
        rt.ExpFileTrack('dr1-mkls0-sa1', os.path.join('TRAIN', 'DR1', 'MKLS0', 'SA1.WAV')),
        rt.ExpFileTrack('dr1-mkls0-sa2', os.path.join('TRAIN', 'DR1', 'MKLS0', 'SA2.WAV')),
        rt.ExpFileTrack('dr1-mrcg0-sx78', os.path.join('TRAIN', 'DR1', 'MRCG0', 'SX78.WAV')),
        rt.ExpFileTrack('dr2-mkjo0-si1517', os.path.join('TRAIN', 'DR2', 'MKJO0', 'SI1517.WAV')),
        rt.ExpFileTrack('dr2-mrfk0-sx176', os.path.join('TRAIN', 'DR2', 'MRFK0', 'SX176.WAV')),
        rt.ExpFileTrack('dr1-fdac1-sa2', os.path.join('TEST', 'DR1', 'FDAC1', 'SA2.WAV')),
        rt.ExpFileTrack('dr1-mjsw0-sa1', os.path.join('TEST', 'DR1', 'MJSW0', 'SA1.WAV')),
        rt.ExpFileTrack('dr1-mjsw0-sx20', os.path.join('TEST', 'DR1', 'MJSW0', 'SX20.WAV')),
        rt.ExpFileTrack('dr2-fpas0-sx224', os.path.join('TEST', 'DR2', 'FPAS0', 'SX224.WAV'))
    ]

    EXPECTED_NUMBER_OF_UTTERANCES = 9
    EXPECTED_UTTERANCES = [
        rt.ExpUtterance('dr1-mkls0-sa1', 'dr1-mkls0-sa1', 'KLS0', 0, float('inf')),
        rt.ExpUtterance('dr1-mkls0-sa2', 'dr1-mkls0-sa2', 'KLS0', 0, float('inf')),
        rt.ExpUtterance('dr1-mrcg0-sx78', 'dr1-mrcg0-sx78', 'RCG0', 0, float('inf')),
        rt.ExpUtterance('dr2-mkjo0-si1517', 'dr2-mkjo0-si1517', 'KJO0', 0, float('inf')),
        rt.ExpUtterance('dr2-mrfk0-sx176', 'dr2-mrfk0-sx176', 'RFK0', 0, float('inf')),
        rt.ExpUtterance('dr1-fdac1-sa2', 'dr1-fdac1-sa2', 'DAC1', 0, float('inf')),
        rt.ExpUtterance('dr1-mjsw0-sa1', 'dr1-mjsw0-sa1', 'JSW0', 0, float('inf')),
        rt.ExpUtterance('dr1-mjsw0-sx20', 'dr1-mjsw0-sx20', 'JSW0', 0, float('inf')),
        rt.ExpUtterance('dr2-fpas0-sx224', 'dr2-fpas0-sx224', 'PAS0', 0, float('inf'))
    ]

    EXPECTED_NUMBER_OF_ISSUERS = 7
    EXPECTED_ISSUERS = [
        rt.ExpSpeaker('DAC1', 1, issuers.Gender.FEMALE, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('JSW0', 2, issuers.Gender.MALE, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('PAS0', 1, issuers.Gender.FEMALE, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('KLS0', 2, issuers.Gender.MALE, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('RCG0', 1, issuers.Gender.MALE, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('KJO0', 1, issuers.Gender.MALE, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('RFK0', 1, issuers.Gender.MALE, issuers.AgeGroup.UNKNOWN, None)
    ]

    RAW_LL = corpus.LL_WORD_TRANSCRIPT_RAW
    WORD_LL = corpus.LL_WORD_TRANSCRIPT
    PHONE_LL = corpus.LL_PHONE_TRANSCRIPT

    EXPECTED_LABEL_LISTS = {
        'dr1-mkls0-sa1': [
            rt.ExpLabelList(RAW_LL, 1),
            rt.ExpLabelList(WORD_LL, 11),
            rt.ExpLabelList(PHONE_LL, 42),
        ],
        'dr1-mkls0-sa2': [
            rt.ExpLabelList(RAW_LL, 1),
            rt.ExpLabelList(WORD_LL, 10),
            rt.ExpLabelList(PHONE_LL, 33),
        ],
        'dr1-mrcg0-sx78': [
            rt.ExpLabelList(RAW_LL, 1),
            rt.ExpLabelList(WORD_LL, 5),
            rt.ExpLabelList(PHONE_LL, 31),
        ],
        'dr2-mkjo0-si1517': [
            rt.ExpLabelList(RAW_LL, 1),
            rt.ExpLabelList(WORD_LL, 7),
            rt.ExpLabelList(PHONE_LL, 42),
        ],
        'dr2-mrfk0-sx176': [
            rt.ExpLabelList(RAW_LL, 1),
        ],
        'dr1-fdac1-sa2': [
            rt.ExpLabelList(RAW_LL, 1),
        ],
        'dr1-mjsw0-sa1': [
            rt.ExpLabelList(RAW_LL, 1),
        ],
        'dr1-mjsw0-sx20': [
            rt.ExpLabelList(RAW_LL, 1),
        ],
        'dr2-fpas0-sx224': [
            rt.ExpLabelList(RAW_LL, 1),
        ],
    }

    EXPECTED_LABELS = {
        'dr1-mkls0-sa1': [
            rt.ExpLabel(RAW_LL, 'She had your dark suit in greasy wash water all year.', 0, float('inf')),
            rt.ExpLabel(WORD_LL, 'she', 0.210625, 0.4275),
            rt.ExpLabel(PHONE_LL, 'h#', 0.0, 0.210625),
        ],
        'dr1-mkls0-sa2': [
            rt.ExpLabel(RAW_LL, 'Don\'t ask me to carry an oily rag like that.', 0, float('inf')),
            rt.ExpLabel(WORD_LL, 'ask', 0.3625, 0.645),
            rt.ExpLabel(PHONE_LL, 'd', 0.209375, 0.244375),
        ],
        'dr1-mrcg0-sx78': [
            rt.ExpLabel(RAW_LL, 'Doctors prescribe drugs too freely.', 0, float('inf')),
            rt.ExpLabel(WORD_LL, 'freely', 1.8575, 2.2898125),
            rt.ExpLabel(PHONE_LL, 't', 0.348125, 0.37375),
        ],
        'dr2-mkjo0-si1517': [
            rt.ExpLabel(RAW_LL, 'Hired, hard lackeys of the warmongering capitalists.', 0, float('inf')),
            rt.ExpLabel(WORD_LL, 'warmongering', 1.461125, 2.18275),
            rt.ExpLabel(PHONE_LL, 'h#', 2.9075, 3.1),
        ],
        'dr2-mrfk0-sx176': [
            rt.ExpLabel(RAW_LL, 'Buying a thoroughbred horse requires intuition and expertise.', 0, float('inf'))
        ],
        'dr1-fdac1-sa2': [
            rt.ExpLabel(RAW_LL, 'Don\'t ask me to carry an oily rag like that.', 0, float('inf')),
        ],
        'dr1-mjsw0-sa1': [
            rt.ExpLabel(RAW_LL, 'She had your dark suit in greasy wash water all year.', 0, float('inf')),
        ],
        'dr1-mjsw0-sx20': [
            rt.ExpLabel(RAW_LL, 'She wore warm, fleecy, woolen overalls.', 0, float('inf')),
        ],
        'dr2-fpas0-sx224': [
            rt.ExpLabel(RAW_LL, 'How good is your endurance?', 0, float('inf')),
        ],
    }

    EXPECTED_NUMBER_OF_SUBVIEWS = 2
    EXPECTED_SUBVIEWS = [
        rt.ExpSubview('TRAIN', [
            'dr1-mkls0-sa1',
            'dr1-mkls0-sa2',
            'dr1-mrcg0-sx78',
            'dr2-mkjo0-si1517',
            'dr2-mrfk0-sx176',
        ]),
        rt.ExpSubview('TEST', [
            'dr1-fdac1-sa2',
            'dr1-mjsw0-sa1',
            'dr1-mjsw0-sx20',
            'dr2-fpas0-sx224',
        ]),
    ]

    def load(self):
        return io.TimitReader().load(self.SAMPLE_PATH)



"""
This module contains all parts needed for using a corpus. Aside the main corpus class
:py:class:`audiomate.Corpus`, there are different loaders in the :py:mod:`audiomate.corpus.io`.
"""

from .base import CorpusView  # noqa: F401

from .corpus import Corpus  # noqa: F401

from audiomate.corpus.subset.subview import Subview  # noqa: F401
from audiomate.corpus.subset.subview import MatchingUtteranceIdxFilter  # noqa: F401

# Definition of common Label-List identifiers
LL_DOMAIN = 'domain'
LL_DOMAIN_MUSIC = 'music'
LL_DOMAIN_SPEECH = 'speech'
LL_DOMAIN_NOISE = 'noise'

LL_WORD_TRANSCRIPT = 'word-transcript'
LL_WORD_TRANSCRIPT_RAW = 'word-transcript-raw'
LL_WORD_TRANSCRIPT_ALIGNED = 'word-transcript-aligned'

LL_PHONE_TRANSCRIPT = 'phone-transcript'
LL_PHONE_TRANSCRIPT_ALIGNED = 'phone-transcript-aligned'

LL_GENRE = 'genre'

LL_SOUND_CLASS = 'sound-class'



import re

from audiomate import annotations
from audiomate.utils import textfile

__TIME_JUNK_PATTERN = re.compile(r'[^0-9.\-]')


def write_label_file(path, entries):
    """
    Writes an audacity label file. Start and end times are in seconds.

    Args:
        path (str): Path to write the file to.
        entries (list): List with entries to write.

    Example::

        >>> data = [
        >>>     [0.0, 0.2, 'sie'],
        >>>     [0.2, 2.2, 'hallo']
        >>> ]
        >>>
        >>> write_label_file('/some/path/to/file.txt', data)
    """

    textfile.write_separated_lines(path, entries, separator='\t')


def write_label_list(path, label_list):
    """
    Writes the given `label_list` to an audacity label file.

    Args:
        path (str): Path to write the file to.
        label_list (audiomate.annotations.LabelList): Label list
    """
    entries = []
    for label in label_list:
        entries.append([label.start, label.end, label.value])

    textfile.write_separated_lines(path, entries, separator='\t')


def read_label_file(path):
    """
    Read the labels from an audacity label file.

    Args:
        path (str): Path to the label file.

    Returns:
        list: List of labels (start [sec], end [sec], label)

    Example::

        >>> read_label_file('/path/to/label/file.txt')
        [
            [0.0, 0.2, 'sie'],
            [0.2, 2.2, 'hallo']
        ]
    """
    labels = []

    for record in textfile.read_separated_lines_generator(path, separator='\t', max_columns=3):
        value = ''

        if len(record) > 2:
            value = str(record[2])

        labels.append([float(_clean_time(record[0])), float(_clean_time(record[1])), value])

    return labels


def read_label_list(path):
    """
    Reads labels from an Audacity label file
    and returns them wrapped in a :py:class:`audiomate.annotations.LabelList`.

    Args:
        path (str): Path to the Audacity label file

    Returns:
        audiomate.annotations.LabelList: Label list containing the labels
    """
    ll = annotations.LabelList()

    for record in read_label_file(path):
        ll.add(annotations.Label(record[2], start=record[0], end=record[1]))

    return ll


def _clean_time(time_str):
    # According to https://en.wikipedia.org/wiki/Decimal_mark, only the comma or the dot are valid as a decimal
    # separator and both may only be used as a decimal separator.
    return re.sub(__TIME_JUNK_PATTERN, '', time_str.replace(',', '.'))



import collections
import copy

import numpy as np

from audiomate import annotations
from audiomate.utils import units


class Utterance(object):
    """
    An utterance defines a sample of audio.
    It is part of a track or can span over the whole track.

    Args:
        idx (str): A unique identifier for the utterance within a dataset.
        track (Track): The track this utterance is belonging to.
        issuer (Issuer): The issuer this utterance was created from.
        start (float): The start of the utterance
                       within the audio track in seconds. (default 0)
        end (float): The end of the utterance within the audio track in seconds.
                     ``inf`` indicates that the utterance ends
                     at the end of the track. (default ``inf``)
        label_lists (LabelList, list): A single or multiple label-lists.

    Attributes:
        label_lists (dict): A dictionary containing label-lists
                            with the label-list-idx as key.
    """

    __slots__ = ['idx', 'track', 'issuer', 'start', 'end', 'label_lists']

    def __init__(self, idx, track, issuer=None, start=0, end=float('inf'), label_lists=None):
        self.idx = idx
        self.track = track
        self.issuer = issuer
        self.start = start
        self.end = end
        self.label_lists = {}

        if label_lists is not None:
            self.set_label_list(label_lists)

        if self.issuer is not None:
            self.issuer.utterances.add(self)

    def __copy__(self):
        return Utterance(
            self.idx,
            self.track,
            issuer=self.issuer,
            start=self.start,
            end=self.end,
            label_lists=list(self.label_lists.values())
        )

    def __deepcopy__(self, memo):
        return Utterance(
            self.idx,
            copy.deepcopy(self.track, memo),
            issuer=copy.deepcopy(self.issuer, memo),
            start=self.start,
            end=self.end,
            label_lists=copy.deepcopy(list(self.label_lists.values()), memo)
        )

    @property
    def end_abs(self):
        """
        Return the absolute end of the utterance relative to the signal.
        """
        if self.end == float('inf'):
            return self.track.duration
        else:
            return self.end

    @property
    def duration(self):
        """
        Return the absolute duration in seconds.
        """
        return self.end_abs - self.start

    def num_samples(self, sr=None):
        """
        Return the number of samples.

        Args:
            sr (int): Calculate the number of samples with the given
                      sampling-rate. If None use the native sampling-rate.

        Returns:
            int: Number of samples
        """
        native_sr = self.sampling_rate
        num_samples = units.seconds_to_sample(self.duration, native_sr)

        if sr is not None:
            ratio = float(sr) / native_sr
            num_samples = int(np.ceil(num_samples * ratio))

        return num_samples

    #
    #   Signal
    #

    def read_samples(self, sr=None, offset=0, duration=None):
        """
        Read the samples of the utterance.

        Args:
            sr (int): If None uses the sampling rate given by the track,
                      otherwise resamples to the given sampling rate.
            offset (float): Offset in seconds to read samples from.
            duration (float): If not ``None`` read only this
                              number of seconds in maximum.

        Returns:
            np.ndarray: A numpy array containing the samples
                        as a floating point (numpy.float32) time series.
        """

        read_duration = self.duration

        if offset > 0 and read_duration is not None:
            read_duration -= offset

        if duration is not None:
            if read_duration is None:
                read_duration = duration
            else:
                read_duration = min(duration, read_duration)

        return self.track.read_samples(
            sr=sr,
            offset=self.start + offset,
            duration=read_duration
        )

    @property
    def sampling_rate(self):
        """
        Return the sampling rate.
        """
        return self.track.sampling_rate

    #
    #   Labels
    #

    def set_label_list(self, label_lists):
        """
        Set the given label-list for this utterance.
        If the label-list-idx is not set, ``default`` is used.
        If there is already a label-list with the given idx,
        it will be overriden.

        Args:
            label_list (LabelList, list): A single or multi. label-lists to add.

        """

        if isinstance(label_lists, annotations.LabelList):
            label_lists = [label_lists]

        for label_list in label_lists:
            if label_list.idx is None:
                label_list.idx = 'default'

            label_list.utterance = self
            self.label_lists[label_list.idx] = label_list

    def all_label_values(self, label_list_ids=None):
        """
        Return a set of all label-values occurring in this utterance.

        Args:
            label_list_ids (list): If not None, only label-values from
                                   label-lists with an id contained in this list
                                   are considered.

        Returns:
             :class:`set`: A set of distinct label-values.
        """
        values = set()

        for label_list in self.label_lists.values():
            if label_list_ids is None or label_list.idx in label_list_ids:
                values = values.union(label_list.label_values())

        return values

    def label_count(self, label_list_ids=None):
        """
        Return a dictionary containing the number of times,
        every label-value in this utterance is occurring.

        Args:
            label_list_ids (list): If not None, only labels from label-lists
                                   with an id contained in this list
                                   are considered.

        Returns:
            dict: A dictionary containing the number of occurrences
                  with the label-value as key.
        """
        count = collections.defaultdict(int)

        for label_list in self.label_lists.values():
            if label_list_ids is None or label_list.idx in label_list_ids:
                for label_value, label_count in label_list.label_count().items():
                    count[label_value] += label_count

        return count

    def all_tokens(self, delimiter=' ', label_list_ids=None):
        """
        Return a list of all tokens occurring in
        one of the labels in the label-lists.

        Args:
            delimiter (str): The delimiter used to split labels into tokens
                             (see :meth:`audiomate.annotations.Label.tokenized`).
            label_list_ids (list): If not None, only labels from label-lists with
                                   an idx contained in this list are considered.

        Returns:
             :class:`set`: A set of distinct tokens.
        """
        tokens = set()

        for label_list in self.label_lists.values():
            if label_list_ids is None or label_list.idx in label_list_ids:
                tokens = tokens.union(label_list.all_tokens(delimiter=delimiter))

        return tokens

    def label_total_duration(self, label_list_ids=None):
        """
        Return a dictionary containing the number of seconds,
        every label-value is occurring in this utterance.

        Args:
            label_list_ids (list): If not None, only labels from label-lists
                                   with an id contained in this
                                   list are considered.

        Returns:
            dict: A dictionary containing the number of seconds
                  with the label-value as key.
        """
        duration = collections.defaultdict(float)

        for label_list in self.label_lists.values():
            if label_list_ids is None or label_list.idx in label_list_ids:
                for label_value, label_duration in label_list.label_total_duration().items():
                    duration[label_value] += label_duration

        return duration

    def split(self, cutting_points, track_relative=False, overlap=0.0):
        """
        Split the utterance into x parts (sub-utterances) and
        return them as new utterances. x is defined by cutting_points
        (``x = len(cutting_points) + 1``).

        By default cutting-points are relative to the start of the utterance.
        For example if an utterance starts at 50s, a cutting-point
        of 10.0 will split the utterance at 60s relative to the track.

        Args:
            cutting_points (list): List of floats defining the times
                                   in seconds where to split the utterance.
            track_relative (bool): If ``True``, cutting-points are relative
                                   to the start of the track. Otherwise they
                                   are relative to the start of the utterance.
            overlap (float): Amount of overlap in seconds. This amount is
                             subtracted from a start-cutting-point,
                             and added to a end-cutting-point.

        Returns:
            list: List of :class:`Utterance`'s.

        Example:

            >>> utt = Utterance('utt-1', 'file-x', start=0.0, end=30.0)
            >>> sub_utts = utt.split([10.0, 20.0])
            >>> len(sub_utts)
            3
            >>> sub_utts[0].start
            0.0
            >>> sub_utts[0].end
            10.0
        """

        if not track_relative:
            cutting_points = [c + self.start for c in cutting_points]

        if len(cutting_points) == 0:
            raise ValueError('At least 1 cutting point is needed!')

        splitted_label_lists = collections.defaultdict(list)

        for idx, label_list in self.label_lists.items():
            label_cutting_points = [x - self.start for x in cutting_points]
            parts = label_list.split(
                label_cutting_points,
                shift_times=True,
                overlap=overlap
            )
            splitted_label_lists[idx] = parts

        # Only consider cutting-points within utterance.
        filtered_cutting_points = []

        for cutting_point in cutting_points:
            if cutting_point > self.start and cutting_point < self.end:
                filtered_cutting_points.append(cutting_point)

        sub_utterances = []

        for index in range(len(filtered_cutting_points) + 1):
            if index == 0:
                sub_start = self.start
            else:
                sub_start = max(self.start, cutting_points[index - 1] - overlap)

            if index >= len(filtered_cutting_points):
                sub_end = self.end
            else:
                sub_end = min(self.end, filtered_cutting_points[index] + overlap)

            new_idx = '{}_{}'.format(self.idx, index)
            new_utt = Utterance(
                new_idx,
                track=self.track,
                issuer=self.issuer,
                start=sub_start,
                end=sub_end
            )

            for parts in splitted_label_lists.values():
                new_utt.set_label_list(parts[index])

            sub_utterances.append(new_utt)

        return sub_utterances



import os
import collections

import audiomate
from audiomate import annotations
from audiomate.corpus import subset
from audiomate.utils import textfile
from . import base
from . import downloader

DOWNLOAD_URL = 'https://github.com/karoldvl/ESC-50/archive/master.zip'
META_FILE_PATH = os.path.join('meta', 'esc50.csv')


class ESC50Downloader(downloader.ArchiveDownloader):
    """
    Downloader for the ESC-50 dataset.

    Args:
        url (str): The url to download the dataset from. If not given the default URL is used.
    """

    def __init__(self, url=None):
        if url is None:
            url = DOWNLOAD_URL

        super(ESC50Downloader, self).__init__(
            url,
            move_files_up=True
        )

    @classmethod
    def type(cls):
        return 'esc-50'


class ESC50Reader(base.CorpusReader):
    """
    Reader for the ESC-50 dataset (Environmental Sound Classification).

    .. seealso::

       `ESC-50 <https://github.com/karoldvl/ESC-50>`_
          Download page
    """

    @classmethod
    def type(cls):
        return 'esc-50'

    def _check_for_missing_files(self, path):
        return []

    def _load(self, path):
        corpus = audiomate.Corpus(path=path)

        meta_data = ESC50Reader.load_meta_data(path)

        folds = collections.defaultdict(list)
        esc10_utt_ids = []

        for record in meta_data:
            file_name = record[0]
            file_id = os.path.splitext(file_name)[0]
            file_path = os.path.abspath(os.path.join(path, 'audio', file_name))
            fold = record[1]
            category = record[3]
            esc10 = record[4]

            corpus.new_file(file_path, file_id)
            utt = corpus.new_utterance(file_id, file_id)
            utt.set_label_list(annotations.LabelList.create_single(category, idx=audiomate.corpus.LL_SOUND_CLASS))

            folds['fold-{}'.format(fold)].append(file_id)

            if esc10 == 'True':
                esc10_utt_ids.append(file_id)

        for fold_id, fold_utt_ids in folds.items():
            fold_filter = subset.MatchingUtteranceIdxFilter(utterance_idxs=set(fold_utt_ids))
            fold_sv = subset.Subview(corpus, filter_criteria=[fold_filter])
            corpus.import_subview(fold_id, fold_sv)

        esc10_filter = subset.MatchingUtteranceIdxFilter(utterance_idxs=set(esc10_utt_ids))
        esc10_sv = subset.Subview(corpus, filter_criteria=[esc10_filter])
        corpus.import_subview('esc-10', esc10_sv)

        return corpus

    @staticmethod
    def load_meta_data(path):
        file_path = os.path.join(path, META_FILE_PATH)
        lines = textfile.read_separated_lines(file_path, separator=',')
        return lines[1:]



import librosa
import numpy as np

from . import base
from . import spectral


class OnsetStrength(base.Computation):
    """
    Compute a spectral flux onset strength envelope.

    Based on http://librosa.github.io/librosa/generated/librosa.onset.onset_strength.html

    Args:
        n_mels (int): Number of mel bands to generate.
    """

    def __init__(self, n_mels=128, parent=None, name=None):
        super(OnsetStrength, self).__init__(left_context=1, right_context=0, parent=parent, name=name)

        self.n_mels = n_mels

    def compute(self, chunk, sampling_rate, corpus=None, utterance=None):
        # Compute mel-spetrogram
        power_spec = np.abs(spectral.stft_from_frames(chunk.data.T)) ** 2
        mel = np.abs(librosa.feature.melspectrogram(S=power_spec, n_mels=self.n_mels, sr=sampling_rate))
        mel_power = librosa.power_to_db(mel)

        # Compute onset strengths
        oenv = librosa.onset.onset_strength(S=mel_power, center=False)

        # Switch dimensions and add dimension to have frames
        oenv = oenv.T.reshape(oenv.shape[0], -1)

        # Remove context
        oenv = oenv[chunk.left_context:oenv.shape[0] - chunk.right_context]

        return oenv



import os

import numpy as np

from audiomate import containers
from audiomate.corpus import subset
from audiomate import feeding
from audiomate.feeding import iterator

import pytest
from tests import resources


@pytest.fixture
def sample_partition_data():
    info = feeding.PartitionInfo()
    info.utt_ids = ['utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5', ]

    data = feeding.PartitionData(info)
    data.utt_data = [
        [np.arange(20).reshape(5, 4), np.arange(20).reshape(5, 4) + 10],
        [np.arange(28).reshape(7, 4), np.arange(28).reshape(7, 4) + 10],
        [np.arange(36).reshape(9, 4), np.arange(36).reshape(9, 4) + 10],
        [np.arange(8).reshape(2, 4), np.arange(8).reshape(2, 4) + 10],
        [np.arange(16).reshape(4, 4), np.arange(16).reshape(4, 4) + 10]
    ]

    return data


class TestDataIterator:

    def test_init_with_utterance_list(self):
        it = feeding.DataIterator(['utt-1', 'utt-2'], [containers.Container('blub')])
        assert set(it.utt_ids) == {'utt-1', 'utt-2'}

    def test_init_with_corpus(self):
        corpus = resources.create_dataset()
        it = feeding.DataIterator(corpus, [containers.Container('blub')])
        assert set(it.utt_ids) == set(corpus.utterances.keys())

    def test_init_with_corpus_view(self):
        corpus = resources.create_dataset()
        subview = subset.Subview(corpus, filter_criteria=[
            subset.MatchingUtteranceIdxFilter(utterance_idxs={'utt-1', 'utt-2', 'utt-4'})
        ])

        it = feeding.DataIterator(subview, [containers.Container('blub')])
        assert set(it.utt_ids) == set(subview.utterances.keys())

    def test_init_throws_error_when_no_container_is_given(self):
        corpus = resources.create_dataset()

        with pytest.raises(ValueError):
            feeding.DataIterator(corpus, [])


class TestMultiFramePartitionData:

    def test_get_utt_regions(self, sample_partition_data):
        frame_data = iterator.MultiFramePartitionData(sample_partition_data, 3, shuffle=False)
        regions = frame_data.get_utt_regions()

        assert regions[0][0] == 0
        assert regions[0][1] == 2

        assert regions[1][0] == 2
        assert regions[1][1] == 3

        assert regions[2][0] == 5
        assert regions[2][1] == 3

        assert regions[3][0] == 8
        assert regions[3][1] == 1

        assert regions[4][0] == 9
        assert regions[4][1] == 2

    def test_get_length(self, sample_partition_data):
        frame_data = iterator.MultiFramePartitionData(sample_partition_data, 3, shuffle=False)
        assert len(frame_data) == 11

    def test_get_item_at_start_of_utterance(self, sample_partition_data):
        frame_data = iterator.MultiFramePartitionData(sample_partition_data, 3, shuffle=False)

        assert len(frame_data[2]) == 2
        assert np.array_equal(frame_data[2][0], np.arange(12).reshape(3, 4))
        assert np.array_equal(frame_data[2][1], np.arange(12).reshape(3, 4) + 10)

    def test_get_item_in_middle_of_utterance(self, sample_partition_data):
        frame_data = iterator.MultiFramePartitionData(sample_partition_data, 3, shuffle=False)

        assert len(frame_data[6]) == 2
        assert np.array_equal(frame_data[6][0], np.arange(12).reshape(3, 4) + 12)
        assert np.array_equal(frame_data[6][1], np.arange(12).reshape(3, 4) + 22)

    def test_get_item_at_end_of_utterance(self, sample_partition_data):
        frame_data = iterator.MultiFramePartitionData(sample_partition_data, 3, shuffle=False)

        assert len(frame_data[8]) == 2
        assert np.array_equal(frame_data[8][0], np.arange(8).reshape(2, 4))
        assert np.array_equal(frame_data[8][1], np.arange(8).reshape(2, 4) + 10)


class TestMultiFrameIterator(object):

    def test_next_emits_no_frames_if_file_is_empty(self, tmpdir):
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        cont = containers.Container(file_path)
        cont.open()

        frames = tuple(iterator.MultiFrameIterator([], [cont], '120', 5))
        assert 0 == len(frames)

    def test_next_emits_no_features_if_data_set_is_empty(self, tmpdir):
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        cont = containers.Container(file_path)
        cont.open()
        cont.set('utt-1', np.array([]))

        frames = tuple(iterator.MultiFrameIterator(['utt-1'], [cont], '120', 5))
        assert 0 == len(frames)

    def test_next_emits_all_features_in_sequential_order(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        cont = containers.Container(file_path)
        cont.open()
        cont.set('utt-1', ds1)
        cont.set('utt-2', ds2)

        frames = tuple(iterator.MultiFrameIterator(['utt-1', 'utt-2'], [cont], '120', 5, shuffle=False))
        assert 2 == len(frames)

        assert np.allclose(([[0.1, 0.1, 0.1, 0.1, 0.1],
                             [0.2, 0.2, 0.2, 0.2, 0.2]]), frames[0][0])
        assert np.allclose(([[0.3, 0.3, 0.3, 0.3, 0.3],
                             [0.4, 0.4, 0.4, 0.4, 0.4],
                             [0.5, 0.5, 0.5, 0.5, 0.5]]), frames[1][0])

    def test_next_emits_all_features_in_random_order(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        ds3 = np.array([[0.6, 0.6, 0.6, 0.6, 0.6]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        cont = containers.Container(file_path)
        cont.open()
        cont.set('utt-1', ds1)
        cont.set('utt-2', ds2)
        cont.set('utt-3', ds3)

        frames = tuple(iterator.MultiFrameIterator(['utt-1', 'utt-2', 'utt-3'], [cont], 120, 2, shuffle=True, seed=6))

        assert 4 == len(frames)

        assert np.allclose(([[0.5, 0.5, 0.5, 0.5, 0.5]]), frames[0][0])
        assert np.allclose(([[0.3, 0.3, 0.3, 0.3, 0.3],
                             [0.4, 0.4, 0.4, 0.4, 0.4]]), frames[1][0])
        assert np.allclose(([[0.6, 0.6, 0.6, 0.6, 0.6]]), frames[2][0])
        assert np.allclose(([[0.1, 0.1, 0.1, 0.1, 0.1],
                             [0.2, 0.2, 0.2, 0.2, 0.2]]), frames[3][0])

    def test_next_emits_features_only_from_included_ds_in_sequential_order(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        ds3 = np.array([[0.6, 0.6, 0.6, 0.6, 0.6], [0.7, 0.7, 0.7, 0.7, 0.7]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        cont = containers.Container(file_path)
        cont.open()
        cont.set('utt-1', ds1)
        cont.set('utt-2', ds2)
        cont.set('utt-3', ds3)

        frames = tuple(iterator.MultiFrameIterator(['utt-1', 'utt-3'], [cont], 120, 2, shuffle=False))

        assert 2 == len(frames)
        assert np.allclose(([[0.1, 0.1, 0.1, 0.1, 0.1],
                             [0.2, 0.2, 0.2, 0.2, 0.2]]), frames[0][0])
        assert np.allclose(([[0.6, 0.6, 0.6, 0.6, 0.6],
                             [0.7, 0.7, 0.7, 0.7, 0.7]]), frames[1][0])

    def test_next_emits_features_only_from_included_ds_in_random_order(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        ds3 = np.array([[0.6, 0.6, 0.6, 0.6, 0.6], [0.7, 0.7, 0.7, 0.7, 0.7]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        cont = containers.Container(file_path)
        cont.open()
        cont.set('utt-1', ds1)
        cont.set('utt-2', ds2)
        cont.set('utt-3', ds3)

        frames = tuple(iterator.MultiFrameIterator(['utt-1', 'utt-3'], [cont], 120, 2, shuffle=True, seed=1))

        assert 2 == len(frames)

        assert np.allclose(([[0.6, 0.6, 0.6, 0.6, 0.6],
                             [0.7, 0.7, 0.7, 0.7, 0.7]]), frames[0][0])
        assert np.allclose(([[0.1, 0.1, 0.1, 0.1, 0.1],
                             [0.2, 0.2, 0.2, 0.2, 0.2]]), frames[1][0])

    def test_next_emits_all_features_if_partition_spans_multiple_data_sets_in_sequential_order(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        ds3 = np.array([[0.6, 0.6, 0.6, 0.6, 0.6], [0.7, 0.7, 0.7, 0.7, 0.7]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        cont = containers.Container(file_path)
        cont.open()
        cont.set('utt-1', ds1)
        cont.set('utt-2', ds2)
        cont.set('utt-3', ds3)

        frames = tuple(iterator.MultiFrameIterator(['utt-1', 'utt-2', 'utt-3'], [cont], 240, 2, shuffle=False))

        assert 4 == len(frames)

        assert np.allclose(([[0.1, 0.1, 0.1, 0.1, 0.1],
                             [0.2, 0.2, 0.2, 0.2, 0.2]]), frames[0][0])
        assert np.allclose(([[0.3, 0.3, 0.3, 0.3, 0.3],
                             [0.4, 0.4, 0.4, 0.4, 0.4]]), frames[1][0])
        assert np.allclose(([[0.5, 0.5, 0.5, 0.5, 0.5]]), frames[2][0])
        assert np.allclose(([[0.6, 0.6, 0.6, 0.6, 0.6],
                             [0.7, 0.7, 0.7, 0.7, 0.7]]), frames[3][0])

    def test_next_emits_all_features_if_partition_spans_multiple_data_sets_in_random_order(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        ds3 = np.array([[0.6, 0.6, 0.6, 0.6, 0.6], [0.7, 0.7, 0.7, 0.7, 0.7]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        cont = containers.Container(file_path)
        cont.open()
        cont.set('utt-1', ds1)
        cont.set('utt-2', ds2)
        cont.set('utt-3', ds3)

        frames = tuple(iterator.MultiFrameIterator(['utt-1', 'utt-2', 'utt-3'], [cont], 240, 2, shuffle=True, seed=12))

        assert 4 == len(frames)

        assert np.allclose(([[0.5, 0.5, 0.5, 0.5, 0.5]]), frames[0][0])
        assert np.allclose(([[0.3, 0.3, 0.3, 0.3, 0.3],
                             [0.4, 0.4, 0.4, 0.4, 0.4]]), frames[1][0])
        assert np.allclose(([[0.1, 0.1, 0.1, 0.1, 0.1],
                             [0.2, 0.2, 0.2, 0.2, 0.2]]), frames[2][0])
        assert np.allclose(([[0.6, 0.6, 0.6, 0.6, 0.6],
                             [0.7, 0.7, 0.7, 0.7, 0.7]]), frames[3][0])

    def test_next_emits_chunks_with_length(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        ds3 = np.array([[0.6, 0.6, 0.6, 0.6, 0.6]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        cont = containers.Container(file_path)
        cont.open()
        cont.set('utt-1', ds1)
        cont.set('utt-2', ds2)
        cont.set('utt-3', ds3)

        frames = tuple(iterator.MultiFrameIterator(['utt-1', 'utt-2', 'utt-3'], [cont], 120, 2, return_length=True,
                                                   shuffle=True, seed=6))

        assert 4 == len(frames)

        assert np.allclose(([[0.5, 0.5, 0.5, 0.5, 0.5]]), frames[0][0])
        assert np.allclose(([[0.3, 0.3, 0.3, 0.3, 0.3],
                             [0.4, 0.4, 0.4, 0.4, 0.4]]), frames[1][0])
        assert np.allclose(([[0.6, 0.6, 0.6, 0.6, 0.6]]), frames[2][0])
        assert np.allclose(([[0.1, 0.1, 0.1, 0.1, 0.1],
                             [0.2, 0.2, 0.2, 0.2, 0.2]]), frames[3][0])

        assert frames[0][1] == 1
        assert frames[1][1] == 2
        assert frames[2][1] == 1
        assert frames[3][1] == 2

    def test_next_emits_chunks_with_padding(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        ds3 = np.array([[0.6, 0.6, 0.6, 0.6, 0.6]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        cont = containers.Container(file_path)
        cont.open()
        cont.set('utt-1', ds1)
        cont.set('utt-2', ds2)
        cont.set('utt-3', ds3)

        frames = tuple(iterator.MultiFrameIterator(['utt-1', 'utt-2', 'utt-3'], [cont], 120, 2, pad=True,
                                                   shuffle=True, seed=6))

        assert 4 == len(frames)

        assert np.allclose(([[0.5, 0.5, 0.5, 0.5, 0.5],
                             [0.0, 0.0, 0.0, 0.0, 0.0]]), frames[0][0])
        assert np.allclose(([[0.3, 0.3, 0.3, 0.3, 0.3],
                             [0.4, 0.4, 0.4, 0.4, 0.4]]), frames[1][0])
        assert np.allclose(([[0.6, 0.6, 0.6, 0.6, 0.6],
                             [0.0, 0.0, 0.0, 0.0, 0.0]]), frames[2][0])
        assert np.allclose(([[0.1, 0.1, 0.1, 0.1, 0.1],
                             [0.2, 0.2, 0.2, 0.2, 0.2]]), frames[3][0])

        assert frames[0][1] == 1
        assert frames[1][1] == 2
        assert frames[2][1] == 1
        assert frames[3][1] == 2


class TestFrameIterator(object):

    def test_next_emits_no_frames_if_file_is_empty(self, tmpdir):
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        cont = containers.Container(file_path)
        cont.open()

        frames = tuple(iterator.FrameIterator([], [cont], 120))
        assert 0 == len(frames)

    def test_next_emits_no_features_if_data_set_is_empty(self, tmpdir):
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        cont = containers.Container(file_path)
        cont.open()
        cont.set('utt-1', np.array([]))

        frames = tuple(iterator.FrameIterator(['utt-1'], [cont], 120))
        assert 0 == len(frames)

    def test_next_emits_all_features_in_sequential_order(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        cont = containers.Container(file_path)
        cont.open()
        cont.set('utt-1', ds1)
        cont.set('utt-2', ds2)

        frames = tuple(iterator.FrameIterator(['utt-1', 'utt-2'], [cont], 120, shuffle=False))
        assert 5 == len(frames)

        assert np.allclose(([0.1, 0.1, 0.1, 0.1, 0.1]), frames[0][0])
        assert np.allclose(([0.2, 0.2, 0.2, 0.2, 0.2]), frames[1][0])
        assert np.allclose(([0.3, 0.3, 0.3, 0.3, 0.3]), frames[2][0])
        assert np.allclose(([0.4, 0.4, 0.4, 0.4, 0.4]), frames[3][0])
        assert np.allclose(([0.5, 0.5, 0.5, 0.5, 0.5]), frames[4][0])

    def test_next_emits_all_features_in_random_order(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        ds3 = np.array([[0.6, 0.6, 0.6, 0.6, 0.6]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        cont = containers.Container(file_path)
        cont.open()
        cont.set('utt-1', ds1)
        cont.set('utt-2', ds2)
        cont.set('utt-3', ds3)

        frames = tuple(iterator.FrameIterator(['utt-1', 'utt-2', 'utt-3'], [cont], 120, shuffle=True, seed=136))

        assert 6 == len(frames)

        assert np.allclose(([0.6, 0.6, 0.6, 0.6, 0.6]), frames[0][0])
        assert np.allclose(([0.1, 0.1, 0.1, 0.1, 0.1]), frames[1][0])
        assert np.allclose(([0.2, 0.2, 0.2, 0.2, 0.2]), frames[2][0])
        assert np.allclose(([0.3, 0.3, 0.3, 0.3, 0.3]), frames[3][0])
        assert np.allclose(([0.5, 0.5, 0.5, 0.5, 0.5]), frames[4][0])
        assert np.allclose(([0.4, 0.4, 0.4, 0.4, 0.4]), frames[5][0])

    def test_next_emits_features_only_from_included_ds_in_sequential_order(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        ds3 = np.array([[0.6, 0.6, 0.6, 0.6, 0.6], [0.7, 0.7, 0.7, 0.7, 0.7]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        cont = containers.Container(file_path)
        cont.open()
        cont.set('utt-1', ds1)
        cont.set('utt-2', ds2)
        cont.set('utt-3', ds3)

        frames = tuple(iterator.FrameIterator(['utt-1', 'utt-3'], [cont], 120, shuffle=False))

        assert 4 == len(frames)

        assert np.allclose(([0.1, 0.1, 0.1, 0.1, 0.1]), frames[0][0])
        assert np.allclose(([0.2, 0.2, 0.2, 0.2, 0.2]), frames[1][0])
        assert np.allclose(([0.6, 0.6, 0.6, 0.6, 0.6]), frames[2][0])
        assert np.allclose(([0.7, 0.7, 0.7, 0.7, 0.7]), frames[3][0])

    def test_next_emits_features_only_from_included_ds_in_random_order(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        ds3 = np.array([[0.6, 0.6, 0.6, 0.6, 0.6], [0.7, 0.7, 0.7, 0.7, 0.7]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        cont = containers.Container(file_path)
        cont.open()
        cont.set('utt-1', ds1)
        cont.set('utt-2', ds2)
        cont.set('utt-3', ds3)

        frames = tuple(iterator.FrameIterator(['utt-1', 'utt-3'], [cont], 120, shuffle=True, seed=1236))

        assert 4 == len(frames)

        assert np.allclose(([0.1, 0.1, 0.1, 0.1, 0.1]), frames[0][0])
        assert np.allclose(([0.2, 0.2, 0.2, 0.2, 0.2]), frames[1][0])
        assert np.allclose(([0.7, 0.7, 0.7, 0.7, 0.7]), frames[2][0])
        assert np.allclose(([0.6, 0.6, 0.6, 0.6, 0.6]), frames[3][0])

    def test_next_emits_all_features_if_partition_spans_multiple_data_sets_in_sequential_order(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        ds3 = np.array([[0.6, 0.6, 0.6, 0.6, 0.6], [0.7, 0.7, 0.7, 0.7, 0.7]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        cont = containers.Container(file_path)
        cont.open()
        cont.set('utt-1', ds1)
        cont.set('utt-2', ds2)
        cont.set('utt-3', ds3)

        frames = tuple(iterator.FrameIterator(['utt-1', 'utt-2', 'utt-3'], [cont], 240, shuffle=False))

        assert 7 == len(frames)

        assert np.allclose(([0.1, 0.1, 0.1, 0.1, 0.1]), frames[0][0])
        assert np.allclose(([0.2, 0.2, 0.2, 0.2, 0.2]), frames[1][0])
        assert np.allclose(([0.3, 0.3, 0.3, 0.3, 0.3]), frames[2][0])
        assert np.allclose(([0.4, 0.4, 0.4, 0.4, 0.4]), frames[3][0])
        assert np.allclose(([0.5, 0.5, 0.5, 0.5, 0.5]), frames[4][0])
        assert np.allclose(([0.6, 0.6, 0.6, 0.6, 0.6]), frames[5][0])
        assert np.allclose(([0.7, 0.7, 0.7, 0.7, 0.7]), frames[6][0])

    def test_next_emits_all_features_if_partition_spans_multiple_data_sets_in_random_order(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        ds3 = np.array([[0.6, 0.6, 0.6, 0.6, 0.6], [0.7, 0.7, 0.7, 0.7, 0.7]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        cont = containers.Container(file_path)
        cont.open()
        cont.set('utt-1', ds1)
        cont.set('utt-2', ds2)
        cont.set('utt-3', ds3)

        frames = tuple(iterator.FrameIterator(['utt-1', 'utt-2', 'utt-3'], [cont], 240, shuffle=True, seed=333))

        assert 7 == len(frames)

        assert np.allclose(([0.5, 0.5, 0.5, 0.5, 0.5]), frames[0][0])
        assert np.allclose(([0.3, 0.3, 0.3, 0.3, 0.3]), frames[1][0])
        assert np.allclose(([0.4, 0.4, 0.4, 0.4, 0.4]), frames[2][0])
        assert np.allclose(([0.2, 0.2, 0.2, 0.2, 0.2]), frames[3][0])
        assert np.allclose(([0.1, 0.1, 0.1, 0.1, 0.1]), frames[4][0])
        assert np.allclose(([0.6, 0.6, 0.6, 0.6, 0.6]), frames[5][0])
        assert np.allclose(([0.7, 0.7, 0.7, 0.7, 0.7]), frames[6][0])
